# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

class LMIDeletedObjectError(Exception):
    """
    Raised, when there is an attempt to access properties on deleted
    :py:class:`LMIInstance` object.
    """

class LMIUnknownParameterError(Exception):
    """
    Raised, when there is a method call issued and unknown method parameter is provided.
    """

class LMIUnknownPropertyError(Exception):
    """
    Raised, when there is an attempt to create instance with unknown property provided.
    """

class LMIMethodCallError(Exception):
    """
    Raised, when an error occurs within method call.
    """

class LMISynchroMethodCallError(Exception):
    """
    Raised, when an error occurs within synchronized method call.
    """

class LMISynchroMethodCallFilterError(Exception):
    """
    Raised, when the LMIShell can not find necessary static filter for synchronous method call.
    """

class LMINoPagerError(Exception):
    """
    Raised, when there is no default pager like less or more.
    """

class LMIIndicationError(Exception):
    """
    Raised, if an error occurs while subscribing to/removing an indication.
    """

class LMIIndicationListenerError(Exception):
    """
    Raised, if there is an error while starting/stopping indication listener.
    """

class LMIHandlerNamePatternError(Exception):
    """
    Raised when the pattern string does not contain minimum replaceable characters.
    """

class LMIFilterError(Exception):
    """
    Raised, when a filter error occurs, mostly when filter object is missing.
    """

class LMIClassNotFound(AttributeError):
    """
    Raised, when trying to access missing class in LMINamespace.

    :param string namespace: namespace name
    :param string classname: class name, which was not found in **namespace**
    """
    def __init__(self, namespace, class_name):
        AttributeError.__init__(self, 'no such class "%s" in "%s" namespace' %
                (class_name, namespace))

class LMINamespaceNotFound(AttributeError):
    """
    Raised, when trying to access not existing namespace from connection
    or namespace object.

    :param string namespace: namespace which was not found
    :param args: other positional arguments
    """
    def __init__(self, namespace, *args):
        AttributeError.__init__(self, 'no such namespace "%s"' %
                "/".join([namespace] + list(args)))




# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem
import signal
import socket
import urlparse
import threading
import collections

from LMIBaseObject import LMIWrapperBaseObject
from LMIBaseClient import LMIBaseClient
from LMIObjectFactory import LMIObjectFactory
from LMIFormatter import LMIMofFormatter
from LMIReturnValue import LMIReturnValue
from LMIConstantValues import LMIConstantValuesParamProp
from LMIConstantValues import LMIConstantValuesMethodReturnType
from LMIIndicationListener import LMIIndicationListener

from LMIJob import lmi_is_job_finished
from LMIJob import lmi_is_job_completed
from LMIJob import lmi_is_job_terminated
from LMIJob import lmi_is_job_killed
from LMIJob import lmi_is_job_exception
from LMIJob import JOB_NOT_FINISHED
from LMIJob import JOB_FINISH_DELAYED
from LMIJob import JOB_FINISH_EARLY

from LMIUtil import LMIPassByRef
from LMIUtil import lmi_raise_or_dump_exception
from LMIUtil import lmi_cast_to_cim
from LMIUtil import lmi_transform_to_lmi

from LMIExceptions import LMIIndicationListenerError
from LMIExceptions import LMIMethodCallError
from LMIExceptions import LMISynchroMethodCallError
from LMIExceptions import LMISynchroMethodCallFilterError
from LMIExceptions import LMIUnknownParameterError
from LMIExceptions import LMIHandlerNamePatternError

class LMISignalHelper(object):
    """
    Helper class, which takes care of signal (de)registration and handling.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(LMISignalHelper, cls).__new__(cls)
            LMISignalHelper.reset(cls._instance)
        return cls._instance

    def reset(self):
        """
        Resets the single instance into default state.
        """
        self._handler_sigint  = None
        self._handler_sigterm = None
        self._signal_handled  = False
        self._instance._callbacks = collections.OrderedDict()

    def signal_attach(self):
        """
        Registers *SIGINT* and *SIGTERM* signals to local handler in which, the flags for
        each signal are modified, if such signal is caught.
        """
        self._signal_handled  = False
        self._handler_sigint  = signal.signal(signal.SIGINT,  LMISignalHelper.__signal_handler)
        self._handler_sigterm = signal.signal(signal.SIGTERM, LMISignalHelper.__signal_handler)

    def signal_detach(self):
        """
        Unregisters *SIGINT* and *SIGTERM* handler and removes all the attached callbacks.
        """
        signal.signal(signal.SIGINT,  self._handler_sigint)
        signal.signal(signal.SIGTERM, self._handler_sigterm)

    def signal_handled(self):
        """
        :returns: True, if any of *SIGINT* or *SIGTERM* has been caught; False otherwise
        """
        return self._signal_handled

    def callback_attach(self, cb_name, cb):
        """
        Registers a callback, which will be called when a *SIGINT* or *SIGTERM* is caught.

        :param string cb_name: callback name
        :param cb: callable object, which takes zero arguments
        """
        self._callbacks[cb_name] = cb

    def callback_detach(self, cb_name):
        """
        Removes a callback from the callback dictionary.

        :param string cb_name: callback name
        """
        self._callbacks.pop(cb_name)

    @staticmethod
    def __signal_handler(signo, frame):
        """
        Signal handler, which is called, when *SIGINT* and *SIGTERM* are sent to
        the LMIShell.

        :param int signo: signal number
        :param frame: -- stack frame

        **NOTE:** see help(signal)
        """
        if signo in (signal.SIGINT, signal.SIGTERM):
            LMISignalHelper._instance._signal_handled = True
        for cb in LMISignalHelper._instance._callbacks.values():
            cb()

class LMIMethod(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`CIMMethod`.

    :param LMIConnection conn: connection object
    :param LMIInstance lmi_instance: :py:class:`LMIInstance` object, on which the method
        call will be issued
    :param string method: method name
    :param bool sync_method: flag indicating, if we are trying to perform a
        synchronous method call
    """
    # 15 seconds sleep timeout for main waiting thread
    _COND_WAIT_TIME = 15
    # Wake count of main thread, when the GetInstance is performed to check,
    # if the job object is present. Prevents infinite waiting for indication
    # delivery. Maximum waiting time, before the GetInstance for job object
    # will be called is: _COND_WAIT_TIME * _COND_WAIT_WAKE_CNT
    _COND_WAIT_WAKE_CNT = 8
    # Default tcp port, where the indications will be delivered.
    # TODO: create a configuration option for the port
    _INDICATION_DESTINATION_PORT = 10240
    # Job classes, which can be used for synchro method calls
    # TODO: create a configuration option for the static filters' classnames
    _INDICATION_JOB_CLASSNAMES = (
        "LMI_StorageJob",
        "LMI_SoftwareInstallationJob",
        "LMI_SoftwareVerificationJob",
        "LMI_NetworkJob"
    )
    # Default namespace where the indication subscriptions, used for synchronous
    # method calls, will be registered.
    _INDICATION_NAMESPACE = "root/interop"
    # When performing a synchronous method call and using the polling method to
    # get a job object status, the sleep time between 2 polls doubles if it is
    # less than _POLLING_ADAPT_MAX_WAITING_TIME.
    _POLLING_ADAPT_MAX_WAITING_TIME = 128

    def __init__(self, conn, lmi_instance, method, sync_method):
        super(LMIMethod, self).__init__(conn)
        self._lmi_instance = lmi_instance
        self._method = lmi_instance._lmi_class._cim_class.methods[method]
        self._sync_method = sync_method
        # Store the constant values as a list. This can consume some time, if computed on demand.
        self._valuemap_parameters_list = [k for (k, v) in self._method.parameters.iteritems() \
            if "ValueMap" in v.qualifiers]
        # For simplicity, we add return value constants to the same list
        if "ValueMap" in self._method.qualifiers:
            self._valuemap_parameters_list.append(self._method.name)

    def __return_synchro_method_call(self, job_inst, job_refresh=True):
        """
        Returns a :py:class:`LMIReturnValue` object with Job output parameters set.

        :param LMIInstance job_inst: job returned from a synchronous method call
        :param bool job_refresh: flag, which indicates, if the ``job_inst``
            needs to be refreshed
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to Job output parameters
        """
        # Adjust return value from the job object
        if job_refresh:
            job_inst.refresh()
        rval = None
        rparams = pywbem.NocaseDict()
        if not job_inst.JobOutParameters is None:
            rparams = pywbem.NocaseDict({
                k: x.value
                for k, x in job_inst.JobOutParameters.properties.iteritems()})
            rval = rparams["__ReturnValue"]
            del rparams["__ReturnValue"]    # NocaseDict has no pop()
        errorstr = ""

        # Is job in exception state? If so, adjust corresponding error string
        # from job.GetError() instance
        if lmi_is_job_exception(job_inst):
            (refreshed, _, errorstr) = job_inst.refresh()
            if not refreshed:
                raise LMISynchroMethodCallError(errorstr)
            (exc_rval, exc_rparams, exc_errorstr) = job_inst.GetError()
            error_inst = exc_rparams.get("error", None)
            if not error_inst:
                raise LMISynchroMethodCallError("Could not get Job error message")
            errorstr = error_inst.Message
        return LMIReturnValue(rval=rval, rparams=rparams, errorstr=errorstr)

    def __handle_synchro_method_call_indication(self, job_inst):
        """
        Handles a synchronous call for asynchronous methods returning a job object. This
        method uses static filters installed by each OpenLMI provider, which is capable of
        using jobs.

        :param LMIInstance job_inst: job object returned from a synchronous
            method call
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to 0 and ``rparams``
            set to job's output parameters
        :raises: :py:exc:`.LMIIndicationListenerError`,
            :py:exc:`.LMISynchroMethodCallError`,
            :py:exc:`.LMISynchroMethodCallFilterError`

        **NOTE:** Static filters' names need to be in format
            "``LMI:<job_class_name>:Changed``"

        """
        def handle_job(ind, cond, job_finished, job_exception):
            """
            :py:class:`LMIListener` handler for synchronous method call, which uses
            indication means of waiting for the job. This function is called, when a job
            changes its state.

            :param threading.Condition cond: condition object used for thread synchronization
            :param LMIPassByRef job_finished: used for synchronization, whether the
                job has finished
            :param LMIPassByRef job_exception: contains an exception object, if any
                exception was raised
            """
            cond.acquire()
            try:
                exp_obj = ind.exported_objects()[0]
                src_inst = exp_obj["SourceInstance"]
                if lmi_is_job_finished(src_inst):
                    # Job has just finished
                    job_finished.value = JOB_FINISH_DELAYED
            except Exception, e:
                # Notify main thread, we are not able to work with such objects
                job_finished.value = JOB_FINISH_DELAYED
                job_exception.value = e
            finally:
                # XXX: Let's be defensive, always notify+release main thread
                cond.notify()
                cond.release()

        # Start indication listener
        cond = threading.Condition()
        job_finished = LMIPassByRef(JOB_NOT_FINISHED)
        job_exception = LMIPassByRef(None)
        # There needs to be a pattern of at least 8 "X" in a row at the end of the indication_name
        indication_name = "synchro-method-call-XXXXXXXX"
        listener = LMIIndicationListener("0.0.0.0", LMIMethod._INDICATION_DESTINATION_PORT)
        indication_name = listener.add_handler(indication_name, handle_job, cond, job_finished,
            job_exception)
        if not listener.start():
            raise LMIIndicationListenerError("Can not start indication listener")

        # Search for necessary static filter
        filter_name = "LMI:%s:Changed" % job_inst.classname
        (cim_filters, _, err) = self._conn._client._get_instances("CIM_IndicationFilter",
            LMIMethod._INDICATION_NAMESPACE, {"Name" : filter_name})
        if not cim_filters:
            listener.stop()
            errorstr = "Can not find proper CIM_IndicationFilter for this method call"
            raise LMISynchroMethodCallFilterError(errorstr)
        cim_filter = cim_filters[0]

        # Create handler object
        netloc = urlparse.urlparse(self._conn.uri).netloc
        if not netloc:
            listener.stop()
            errorstr = "Can not determine netloc from client's uri"
            raise LMISynchroMethodCallError(errorstr)
        netloc = netloc.split(":")[0]
        # NOTE: This will work only on a local area network. Complicated networks may require
        # additional configuration to make this work. See LMIMethod() and PreferPolling.
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect((netloc, LMIMethod._INDICATION_DESTINATION_PORT))
        except socket.gaierror, e:
            listener.stop()
            errorstr = "Can not determine IP address of this machine"
            raise LMISynchroMethodCallError(errorstr)
        destination = s.getsockname()[0]
        s.close()
        # NOTE: For now, we are using insecure HTTP as a transport protocol.
        # TODO: Switch to HTTPS
        cim_handler_props = {
            "Name" : indication_name,
            "Destination" : "http://%s:%d/%s" % (destination,
                LMIMethod._INDICATION_DESTINATION_PORT, indication_name),
        }
        (cim_handler, _, err) = self._conn._client._create_instance("CIM_IndicationHandlerCIMXML",
            LMIMethod._INDICATION_NAMESPACE, cim_handler_props)
        if not cim_handler:
            listener.stop()
            errorstr = "Can not create CIM_IndicationHandlerCIMXML object"
            raise LMISynchroMethodCallError(errorstr)

        # Create indication subscription object
        cim_subscription_props = {
            "Filter" : cim_filter.path,
            "Handler" : cim_handler.path
        }
        (cim_subscription, _, err) = self._conn._client._create_instance("CIM_IndicationSubscription",
            LMIMethod._INDICATION_NAMESPACE, cim_subscription_props)
        if not cim_subscription:
            self._conn._client._delete_instance(cim_handler)
            listener.stop()
            errorstr = "Can not create CIM_IndicationSubscription object"
            raise LMISynchroMethodCallError(errorstr)

        # Check, if the job is not already in finished state,
        # while we were subscribing for the indications
        job_inst.refresh()
        if lmi_is_job_finished(job_inst):
            job_finished.value = JOB_FINISH_EARLY

        # Register signal callback for SIGINT, SIGTERM with callback,
        # which awakes waiting thread for immediate return.
        LMISignalHelper().callback_attach("indication", lambda: LMIMethod.__wake(cond))
        LMISignalHelper().signal_attach()

        # Wait for the job to finish
        wake_cnt = 0
        cond.acquire()
        while not LMISignalHelper().signal_handled() and \
                not job_finished.value and \
                not lmi_is_job_finished(job_inst):
            cond.wait(LMIMethod._COND_WAIT_TIME)
            wake_cnt += 1
            # XXX: threading.Condition.wait() does not inform about timeout or being awaken by
            # notify call. There is a counting to 4 sleep cycles before we actually check for
            # job status manually. This number can be increased, so we rely more on indications,
            # rather then on manual polling.
            if wake_cnt >= LMIMethod._COND_WAIT_WAKE_CNT and \
                    not job_finished.value:
                wake_cnt = 0
                try:
                    (refreshed, _, errorstr) = job_inst.refresh()
                except pywbem.cim_operations.CIMError, e:
                    job_exception.value = e
                    break
                if not refreshed:
                    job_exception.value = LMISynchroMethodCallError(errorstr)
                    break

        # Unregister signal handler
        LMISignalHelper().signal_detach()
        LMISignalHelper().callback_detach("indication")

        cond.release()

        # Cleanup
        listener.stop()
        self._conn._client._delete_instance(cim_subscription.path)
        self._conn._client._delete_instance(cim_handler.path)
        if job_exception.value:
            raise job_exception.value
        if LMISignalHelper().signal_handled() and not job_finished.value:
            # We got SIGINT or SIGTERM, when waiting for the job, cancelling the job
            sys.stderr.write("Cancelling a job '%s'\n" % job_inst.Name)
            job_inst.RequestStateChange(
                RequestedState=job_inst.RequestStateChange.RequestedStateValues.Terminate
            )
            return LMIReturnValue(rval=None)

        # Return the job return values, refresh the job_inst object, if we got notified
        # about the job finish state by indication.
        return self.__return_synchro_method_call(job_inst, job_finished.value == JOB_FINISH_DELAYED)

    def __handle_synchro_method_call_polling(self, job_inst):
        """
        Handles a synchronous call for asynchronous methods returning a job object.
        This call uses polling method to wait for the job to finish.

        :param LMIInstance job_inst: job object
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to 0 and ``rparams``
            set to job's output parameters.
        :raises: :py:exc:`.LMISynchroMethodCallError`
        """

        # Register signal callback for SIGINT, SIGTERM with callback,
        # which awakes waiting thread for immediate return.
        LMISignalHelper().callback_attach("polling", lambda: LMIMethod.__wake(cond))
        LMISignalHelper().signal_attach()

        cond = threading.Condition()
        cond.acquire()

        job_exception = None

        try:
            sleep_time = 1
            while not LMISignalHelper().signal_handled() and \
                    not lmi_is_job_finished(job_inst):
                # Sleep, a bit longer in every iteration
                cond.wait(sleep_time)
                if sleep_time < LMIMethod._POLLING_ADAPT_MAX_WAITING_TIME:
                    sleep_time *= 2
                (refreshed, _, errorstr) = job_inst.refresh()
                if not refreshed:
                    job_exception = LMISynchroMethodCallError(errorstr)
                    break
        except pywbem.cim_operations.CIMError, e:
            job_exception = LMISynchroMethodCallError(e.message)
        finally:
            cond.release()

        # Unregister signal handler and callback
        LMISignalHelper().signal_detach()
        LMISignalHelper().callback_detach("polling")

        if LMISignalHelper().signal_handled() and not lmi_is_job_finished(job_inst):
            # We got SIGINT or SIGTERM, when waiting for the job, cancelling the job
            sys.stderr.write("Cancelling a job '%s'\n" % job_inst.Name)
            job_inst.RequestStateChange(
                RequestedState=job_inst.RequestStateChange.RequestedStateValues.Terminate
            )
            return LMIReturnValue(rval=None)

        if not job_exception is None:
            raise job_exception

        # Return the job return values. No need to refresh the job instance, we already
        # have a "fresh" one.
        return self.__return_synchro_method_call(job_inst, False)

    def __call__(self, method_args=None, **kwargs):
        """
        Perform a method call.

        Method arguments are preferably passed by dictionary (parameter : value). Using
        former means of passing arguments, by keyword arguments, to a method call works
        too.

        If performing a synchronous method call, passing PreferPolling can be used
        to select which method should be used -- either subscribing to an indication
        or polling method.

        :param dictionary method_args: method arguments
        :param dictionary kwargs: keyword method arguments (``method_args`` preffered)

            * **RefreshInstance** (*bool*) flag, which tells the LMIShell, whether
              the instance should be refreshed after a method call. Default
              value is False.

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to method's return
            value; ``rparams`` set to method's return parameters; ``errorstr`` set
            method's error string in case of failure
        :raises: :py:exc:`.LMIUnknownParameterError`,
            :py:exc:`.LMIMethodCallError`
            :py:exc:`.LMISynchroMethodCallError`

        **Usage:** :ref:`instances_methods`.
        """
        synchro_method_polling = kwargs.pop("PreferPolling", False)
        refresh_requested = kwargs.pop("RefreshInstance", False)
        if method_args is None:
            method_args = {}
        method_args.update(kwargs)
        for (param, value) in method_args.iteritems():
            if param in self._method.parameters:
                if isinstance(value, LMIObjectFactory().LMIInstanceName):
                    value = value.wrapped_object
                t = self._method.parameters[param].type
                method_args[param] = lmi_cast_to_cim(t, value)
            else:
                # NOTE: maybe we could check for pywbem type and not to exit prematurely
                errorstr = "Unknown parameter '%s' supplied for method '%s'" % (param, self._method.name)
                lmi_raise_or_dump_exception(LMIUnknownParameterError(errorstr))
                return LMIReturnValue(rval=-1, errorstr=errorstr)
        (rval, call_rparams, call_errorstr) = self._conn._client._call_method_raw(self._lmi_instance,
            self._method.name, **method_args)
        rval = lmi_transform_to_lmi(self._conn, rval)
        if call_rparams:
            call_rparams = lmi_transform_to_lmi(self._conn, call_rparams)
            if not call_rparams:
                # NOTE: this is wrong! What should we do?
                errorstr = "Could not perform CIM -> LMI object transformation"
                lmi_raise_or_dump_exception(LMIMethodCallError(errorstr))
                return LMIReturnValue(rval=rval, errorstr=errorstr)
            # Check if we can perform synchronous method call
            job = call_rparams.get("job", None)
            can_perform_sync_call = job.classname in LMIMethod._INDICATION_JOB_CLASSNAMES if job else False
            if self._sync_method and job and can_perform_sync_call:
                # Synchronous method calls
                job_inst = call_rparams['job'].to_instance()
                # At first, try to wait for the call to finish by subscribing to an indication
                handled_by_indication = False
                if not synchro_method_polling:
                    try:
                        (rval, call_rparams, call_errorstr) = self.__handle_synchro_method_call_indication(job_inst)
                        handled_by_indication = True
                    except pywbem.CIMError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.args[1])
                    except LMISynchroMethodCallError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.message)
                    # Fall through, try to handle the synchro call by polling
                    except LMIHandlerNamePatternError, e:
                        handled_by_indication = False
                    except LMISynchroMethodCallFilterError, e:
                        handled_by_indication = False
                    except LMIIndicationListenerError, e:
                        handled_by_indication = False
                if not handled_by_indication:
                    # Executed, when LMIListener can not be started
                    try:
                        (rval, call_rparams, call_errorstr) = self.__handle_synchro_method_call_polling(job_inst)
                    except LMISynchroMethodCallError, e:
                        lmi_raise_or_dump_exception(e)
                        return LMIReturnValue(rval=-1, errorstr=e.message)
                call_rparams = lmi_transform_to_lmi(self._conn, call_rparams)
        if refresh_requested and not self._lmi_instance.refresh():
            # NOTE: this is wrong! What should we do?
            errorstr = "Could not update an LMI object after a method call"
            lmi_raise_or_dump_exception(LMIMethodCallError(errorstr))
            return LMIReturnValue(rval=rval, errorstr=errorstr)
        return LMIReturnValue(rval=rval, rparams=call_rparams, errorstr=call_errorstr)

    def __getattr__(self, name):
        """
        Returns either a class member, or a constant value.

        :param string name: class member, or the constant value name
        """
        if name in self.__dict__:
            return self.__dict__[name]
        if name.endswith("Values"):
            parameter_name = name[:-6]
            if parameter_name in self._method.parameters:
                return LMIConstantValuesParamProp(self._method.parameters[parameter_name])
            elif parameter_name == self._method.name:
                return LMIConstantValuesMethodReturnType(self._method)
        raise AttributeError(name)

    @staticmethod
    def __wake(cond):
        """
        Helper function used for manual :py:attr:`threading.Condition` wakeup.

        :param threading.Condition cond: condition object
        """
        cond.acquire()
        cond.notify()
        cond.release()

    @property
    def return_type(self):
        """
        :returns: string of the method call's return type
        """
        return self._method.return_type

    def tomof(self):
        """
        Prints out a message with MOF representation of :py:class:`CIMMethod`. If the
        LMIShell is run in a interactive mode, the output will be redirected to a pager
        set by environment variable :envvar:`PAGER`. If there is not :envvar:`PAGER` set,
        less or more will be used as a fall-back.
        """
        LMIMofFormatter(self._method.tomof()).fancy_format(self._conn._client.interactive)

    def valuemap_parameters(self):
        """
        :returns: list of strings of the constant names
        """
        return self._valuemap_parameters_list

    def print_valuemap_parameters(self):
        """
        Prints out the list of strings of constant names.
        """
        for i in self._valuemap_parameters_list:
            sys.stdout.write("%s\n" % i)

    def parameters(self):
        """
        :returns: list of strings of :py:class:`CIMMethod`'s parameters
        """
        return self._method.parameters

    def print_parameters(self):
        """
        Prints out :py:class:`CIMMethod`'s parameters.
        """
        for (param, value) in self._method.parameters.iteritems():
            sys.stdout.write("%s %s%s\n" % (value.type, param, "[]" if value.is_array else ""))

    @property
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`CIMmethod` object
        """
        return self._method



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import sys
import atexit
import pywbem
import readline
import urlparse

import OpenSSL.SSL

from LMIBaseClient import LMIBaseClient
from LMIShellClient import LMIShellClient
from LMINamespace import LMINamespaceRoot
from LMIInstance import LMIInstance
from LMIReturnValue import LMIReturnValue
from LMISubscription import LMISubscription

from LMIExceptions import LMIIndicationError
from LMIExceptions import LMINamespaceNotFound

from LMIUtil import lmi_raise_or_dump_exception
from LMIUtil import lmi_get_use_exceptions
from LMIUtil import lmi_set_use_exceptions

def __lmi_raw_input(prompt, use_echo=True):
    """
    Reads a string from the standard input.

    :param string prompt: input prompt
    :param bool use_echo: whether to echo the input on the command line
    :returns: input string
    :raises: :py:exc:`EOFError`
    """
    if not sys.stdout.isatty() and sys.stderr.isatty():
        # read the input with prompt printed to stderr
        def get_input(prompt):
            sys.stderr.write(prompt)
            return raw_input()
        stream = sys.stderr
    else:
        # read the input with prompt printed to stdout
        # NOTE: raw_input uses stdout only if the readline module is imported
        get_input = raw_input
        stream = sys.stdout
    if not sys.stderr.isatty() and not sys.stdout.isatty():
        LOG.warn('both stdout and stderr are detached from terminal,'
            ' using stdout for prompt')
    if not use_echo:
        os.system("stty -echo")
    try:
        result = get_input(prompt)
    except EOFError, e:
        if not use_echo:
            os.system("stty echo")
        stream.write("\n")
        return None
    except KeyboardInterrupt, e:
        if not use_echo:
            os.system("stty echo")
        raise e
    if not use_echo:
        os.system("stty echo")
        stream.write("\n")
    if result:
        cur_hist_len = readline.get_current_history_length()
        if cur_hist_len > 1:
            readline.remove_history_item(cur_hist_len - 1)

    return result

def connect(uri, username="", password="", **kwargs):
    """
    Creates a connection object with provided URI and credentials.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be performed
    :param string password: user's password

    :param bool interactive: flag indicating, if the LMIShell client is running in the
        interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use cache for
        :py:class:`CIMClass` objects. This saves lot's of communication, if there are
        :func:`EnumerateInstances` and :func:`EnumerateClasses` intrinsic methods often
        issued. Default value is True.
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: flag indicating, whether a server side certificate
        needs to be verified, if SSL used; default value is True.
    :param string prompt_prefix: username and password prompt prefix in case the user is
        asked for credentials. Default value is empty string.
    :returns: :py:class:`LMIConnection` object or None, if LMIShell does not use
        exceptions
    :raises: :py:exc:`pywbem.AuthError`

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`,
      :py:meth:`.LMIClass.doc`, :py:meth:`.LMIInstance.tomof` and
      :py:meth:`.LMIMethod.tomof`

    **Usage:** :ref:`startup_connection`.
    """
    # Set remaining arguments
    interactive = kwargs.pop("interactive", False)
    use_cache = kwargs.pop("use_cache", True)
    key_file = kwargs.pop("key_file", None)
    cert_file = kwargs.pop("cert_file", None)
    verify_server_cert = kwargs.pop("verify_server_cert", True)
    prompt_prefix = kwargs.pop("prompt_prefix", "")
    if kwargs:
        raise TypeError("connect() got an unexpected keyword arguments: %s" % ", ".join(kwargs.keys()))

    connection = None
    netloc = urlparse.urlparse(uri).netloc
    destination = netloc if netloc else uri
    if os.getuid() == 0 and destination in ("localhost", "127.0.0.1", "::1") and \
            os.path.exists("/var/run/tog-pegasus/cimxml.socket") and \
            not username and not password:
        connection = LMIConnection(uri, None, None, interactive=interactive,
            use_cache=use_cache, conn_type=LMIBaseClient.CONN_TYPE_PEGASUS_UDS,
            verify_server_cert=verify_server_cert)
        if not connection.verify_credentials():
            connection = None
    if connection is None:
        if interactive and not key_file and not cert_file:
            try:
                if not username:
                    username = __lmi_raw_input(prompt_prefix+"username: ", True)
                if not password:
                    password = __lmi_raw_input(prompt_prefix+"password: ", False)
            except KeyboardInterrupt, e:
                sys.stdout.write("\n")
                return None
        connection = LMIConnection(uri, username, password, interactive=interactive,
            use_cache=use_cache, conn_type=LMIBaseClient.CONN_TYPE_WBEM,
            key_file=key_file, cert_file=cert_file, verify_server_cert=verify_server_cert)
        if not connection.verify_credentials():
            return None
    return connection

class LMIConnection(object):
    """
    Class representing a connection object. Each desired connection to separate CIMOM
    should have its own connection object created. This class provides an entry point to
    the namespace/classes/instances/methods hierarchy present in the LMIShell.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be performed
    :param string password: user's password
    :param bool interactive: flag indicating, if the LMIShell client is running
      in the interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use cache for
        CIMClass objects. This saves lot's of communication, if there are
        :func:`EnumerateInstances` and :func:`EnumerateClasses` intrinsic methods often
        issued. Default value is True.
    :param conn_type: type of connection; can be of 2 values:

        * :py:attr:`LMIBaseClient.CONN_TYPE_WBEM` -- WBEM connection,
        * :py:attr:`LMIBaseClient.CONN_TYPE_PEGASUS_UDS` -- applicable only for
          Tog-Pegasus CIMOM, it uses Unix socket for the connection; default value is
          :py:attr:`LMIBaseClient.CONN_TYPE_WBEM`
    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: flag indicating, whether a server side certificate
        needs to be verified, if SSL used; default value is True

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`, :py:meth:`.LMIClass.doc`,
      :py:meth:`.LMIInstance.tomof` and :py:meth:`.LMIMethod.tomof`
    """
    def __init__(self, uri, username="", password="", **kwargs):
        # Set remaining arguments
        interactive = kwargs.pop("interactive", False)
        use_cache = kwargs.pop("use_cache", True)
        conn_type = kwargs.pop("conn_type", LMIBaseClient.CONN_TYPE_WBEM)
        key_file = kwargs.pop("key_file", None)
        cert_file = kwargs.pop("cert_file", None)
        verify_server_cert = kwargs.pop("verify_server_cert", True)
        if kwargs:
            raise TypeError("__init__() got an unexpected keyword arguments: %s" % ", ".join(kwargs.keys()))

        self._client = LMIShellClient(uri, username, password, interactive=interactive,
            use_cache=use_cache, conn_type=conn_type, key_file=key_file, cert_file=cert_file,
            verify_server_cert=verify_server_cert)
        self._indications = {}
        # Register LMIConnection.__unsubscribe_all_indications() to be called at LMIShell's exit.
        atexit.register(lambda: self.__unsubscribe_all_indications())

    def __repr__(self):
        """
        :returns: pretty string for the object.
        """
        return "%s(URI='%s', user='%s'...)" % (self.__class__.__name__,
            self._client.uri, self._client.username)

    @property
    def uri(self):
        """
        :returns: URI of the CIMOM
        :rtype: string
        """
        return self._client.uri

    @property
    def namespaces(self):
        """
        :returns: list of all available namespaces

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        return ["root"]

    @property
    def root(self):
        """
        :returns: :py:class:`.LMINamespaceRoot` object for *root* namespace
        """
        return LMINamespaceRoot(self)

    def print_namespaces(self):
        """
        Prints out all available namespaces.
        """
        sys.stdout.write("root\n")

    def get_namespace(self, namespace):
        """
        :param string namespace: namespace path (eg. `root/cimv2`)
        :returns: :py:class:`LMINamespace` object
        :raises: :py:exc:`.LMINamespaceNotFound`
        """
        def get_namespace_priv(namespace, namespace_path):
            if not namespace_path:
                return namespace
            ns = namespace_path.pop(0)
            return get_namespace_priv(getattr(namespace, ns), namespace_path)

        namespace_path = namespace.split("/")
        ns = namespace_path.pop(0)
        if not ns in self.namespaces:
            raise LMINamespaceNotFound(ns)
        return get_namespace_priv(getattr(self, ns), namespace_path)

    def clear_cache(self):
        """
        Clears the cache.
        """
        self._client._cache.clear()

    def use_cache(self, active=True):
        """
        Sets a bool flag, which defines, if the LMIShell should use a cache.

        :param bool active: whether the LMIShell's cache should be used
        """
        self._client._cache.active = active

    def verify_credentials(self):
        """
        Verifies credentials by performing a "dummy" :func:`GetClass` call on
        "SomeNonExistingClass". Provided credentials are OK, if the LMIShell
        obtains :py:exc:`pywbem.CIMError` exception with the flag
        ``CIM_ERR_NOT_FOUND`` set. Otherwise, the should receive
        :py:exc:`pywbem.AuthError`.

        :returns: True if provided credentials are OK; False otherwise
        """
        try:
            use_exceptions = lmi_get_use_exceptions()
            lmi_set_use_exceptions(True)
            try:
                self._client._get_class("SomeNonExistingClass")
            except:
                raise
            finally:
                lmi_set_use_exceptions(use_exceptions)
        except pywbem.cim_operations.CIMError, e:
            if e.args[0] == pywbem.cim_constants.CIM_ERR_NOT_FOUND:
                return True
            lmi_raise_or_dump_exception(e)
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
        except OpenSSL.SSL.Error, e:
            lmi_raise_or_dump_exception(e)
        return False

    def subscribe_indication(self, **kwargs):
        """
        Subscribes to an indication. Indication is formed by 3 objects, where 2 of them
        (filter and handler) can be provided, if the LMIShell should not create those 2 by itself.

        **NOTE:** Currently the call registers :py:mod:`atexit` hook, which auto-deletes
        all subscribed indications by the LMIShell.

        :param dictionary kwargs: parameters for the indication subscription

            * **Filter** (*LMIInstance*) -- if provided, the
              :py:class:`LMIInstance` object will be used instead of creating a new one;
              optional
            * **Handler** (*LMIInstance*) -- if provided, the :py:class:`LMIInstance`
              object will be used instead of creating a new one; optional
            * **Query** (*string*) -- string containing a query for the indications filtering
            * **QueryLanguage** (*string*) -- query language; eg. `WQL`, or `DMTF:CQL`
            * **Name** (*string*) -- indication name
            * **CreationNamespace** (*string*) -- creation namespace
            * **SubscriptionCreationClassName** (*string*) -- subscription object class
              name
            * **Permanent** (*bool*) -- whether to preserve the created subscription on
              LMIShell's quit. Default value is False.
            * **FilterCreationClassName** (*string*) -- creation class name of the filter
              object
            * **FilterSystemCreationClassName** (*string*) -- system creation class name
              of the filter object
            * **FilterSourceNamespace** (*string*) -- local namespace where the
              indications originate
            * **HandlerCreationClassName** (*string*) -- creation class name of the
              handler object
            * **HandlerSystemCreationClassName** (*string*) -- system creation name of the
              handler object
            * **Destination** (*string*) -- destination URI, where the indications should
              be delivered

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to True, if
            indication was subscribed; False otherwise. If a error occurs, ``errorstr`` is
            set to appropriate error string.
        """
        try:
            cim_filter_provided = "Filter" in kwargs
            if cim_filter_provided:
                filt = kwargs["Filter"]
                cim_filter = None
                if isinstance(filt, LMIInstance):
                    cim_filter = filt._cim_instance
                elif isinstance(filt, pywbem.CIMInstance):
                    cim_filter = filt
                else:
                    errorstr = "Filter argument accepts instances of CIMInstance or LMIInstance"
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            else:
                cim_filter_props = {
                    "CreationClassName" : kwargs["FilterCreationClassName"],
                    "SystemCreationClassName" : kwargs["FilterSystemCreationClassName"],
                    "SourceNamespace" : kwargs["FilterSourceNamespace"],
                    "SystemName" : self._client.uri,
                    "Query" : kwargs["Query"],
                    "QueryLanguage" : kwargs["QueryLanguage"],
                    "Name" : kwargs["Name"] + "-filter"
                }
                (cim_filter, _, errorstr) = self._client._create_instance(
                    kwargs["FilterCreationClassName"], kwargs["CreationNamespace"],
                    cim_filter_props)
                if not cim_filter:
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            cim_handler_provided = "Handler" in kwargs
            if cim_handler_provided:
                cim_handler = kwargs["Handler"]._cim_instance
            else:
                cim_handler_props = {
                    "CreationClassName" : kwargs["HandlerCreationClassName"],
                    "SystemCreationClassName" : kwargs["HandlerSystemCreationClassName"],
                    "SystemName" : self._client.uri,
                    "Destination" : kwargs["Destination"] + "/" + kwargs["Name"],
                    "Name" : kwargs["Name"] + "-handler"
                }
                (cim_handler, _, errorstr) = self._client._create_instance(
                    kwargs["HandlerCreationClassName"], kwargs["CreationNamespace"],
                    cim_handler_props)
                if not cim_handler:
                    if not "Filter" in kwargs:
                        self._client._delete_instance(cim_filter.path)
                    lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                    return LMIReturnValue(rval=False, errorstr=errorstr)
            cim_subscription_props = {
                "Filter" : cim_filter.path,
                "Handler" : cim_handler.path
            }
            (cim_subscription, _, errorstr) = self._client._create_instance(
                kwargs["SubscriptionCreationClassName"], kwargs["CreationNamespace"],
                cim_subscription_props)
            if not cim_subscription:
                if not "Filter" in kwargs:
                    self._client._delete_instance(cim_filter.path)
                if not "Handler" in kwargs:
                    self._client._delete_instance(cim_handler.path)
                lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
                return LMIReturnValue(rval=False, errorstr=errorstr)
            # XXX: Should we auto-delete all the indications?
            permanent = kwargs.get("Permanent", False)
            self._indications[kwargs["Name"]] = LMISubscription(
                self._client,
                (cim_filter, not cim_filter_provided),
                (cim_handler, not cim_handler_provided),
                cim_subscription,
                permanent)
        except KeyError, e:
            errorstr = "Not all necessary parameters provided, missing: %s" % e
            lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
            return LMIReturnValue(rval=False, errorstr=errorstr)
        return LMIReturnValue(rval=True)

    def unsubscribe_indication(self, name):
        """
        Unsubscribes an indication.

        :param string name: indication name
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to True, if
            unsubscribed; False otherwise
        """
        if not name in self._indications:
            errorstr = "No such indication"
            lmi_raise_or_dump_exception(LMIIndicationError(errorstr))
            return LMIReturnValue(rval=False, errorstr=errorstr)
        indication = self._indications.pop(name)
        indication.delete()
        return LMIReturnValue(rval=True)

    def __unsubscribe_all_indications(self):
        """
        Unsubscribes all the indications, which were not marked as Permanent.
        """
        def delete_subscription(subscription):
            if subscription.permanent:
                return
            subscription.delete()
        map(lambda obj: delete_subscription(obj), self._indications.values())
        self._indications = {}

    def unsubscribe_all_indications(self):
        """
        Unsubscribes all the indications. This call ignores *Permanent* flag, which may be
        provided in :py:meth:`.LMIConnection.subscribe_indication`, and deletes all the
        subscribed indications.
        """
        map(lambda obj: obj.delete(), self._indications.values())
        self._indications = {}

    def print_subscribed_indications(self):
        """
        Prints out all the subscribed indications.
        """
        for i in self._indications.keys():
            sys.stdout.write("%s\n" % i)

    def subscribed_indications(self):
        """
        :returns: list of all the subscribed indications
        """
        return self._indications.keys()



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
LMIShell
"""

from LMIShellVersion import __version__

from LMIExceptions import *
from LMIUtil import LMIPassByRef
from LMIIndication import LMIIndication
from LMIIndicationListener import LMIIndicationListener
from LMIReturnValue import LMIReturnValue
from LMIBaseClient import LMIBaseClient
from LMIShellClient import LMIShellClient
from LMIShellConfig import LMIShellConfig
from LMIShellOptions import LMIShellOptions
from LMISubscription import LMISubscription
from LMINamespace import LMINamespace
from LMINamespace import LMINamespaceRoot
from LMIConstantValues import LMIConstantValues
from LMIClass import LMIClass
from LMIInstanceName import LMIInstanceName
from LMIMethod import LMIMethod
from LMIInstance import LMIInstance
from LMIConnection import LMIConnection

from LMIConnection import connect

# Register LMI wrapper classes into a LMIObjectFactory  to avoid circular dependency
# between those classes.
from LMIObjectFactory import LMIObjectFactory
LMIObjectFactory().register(LMINamespace)
LMIObjectFactory().register(LMINamespaceRoot)
LMIObjectFactory().register(LMIClass)
LMIObjectFactory().register(LMIInstance)
LMIObjectFactory().register(LMIInstanceName)
LMIObjectFactory().register(LMIMethod)

# LMIConsole depends on LMIObjectFactory and needs to be imported after
# LMIObjectFactory has all wrapper classes registered.
from LMIConsole import LMIConsole



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem

from LMIBaseObject import LMIWrapperBaseObject
from LMIFormatter import LMIXmlFormatter
from LMIFormatter import LMIMofFormatter
from LMIObjectFactory import LMIObjectFactory
from LMIReturnValue import LMIReturnValue

from LMIDecorators import lmi_possibly_deleted
from LMIDecorators import lmi_return_expr_if_fail
from LMIDecorators import lmi_return_val_if_fail
from LMIDecorators import lmi_return_if_fail

from LMIUtil import lmi_cast_to_cim
from LMIUtil import lmi_wrap_cim_instance
from LMIUtil import lmi_wrap_cim_instance_name
from LMIUtil import lmi_wrap_cim_method

class LMIInstance(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`CIMInstance`.

    :param LMIConnection conn: connection object
    :param LMIClass lmi_class: wrapped creation class of the instance
    :param CIMInstance cim_instance: wrapped object
    """
    def __init__(self, conn, lmi_class, cim_instance):
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        if isinstance(cim_instance, LMIInstance):
            cim_instance = cim_instance.wrapped_object
        self.__dict__["_deleted"] = False
        self.__dict__["_cim_instance"] = cim_instance
        self.__dict__["_lmi_class"] = lmi_class
        super(LMIInstance, self).__init__(conn)

    @lmi_possibly_deleted(False)
    def __contains__(self, key):
        """
        Returns True, if the specified key is present in the properties, False otherwise.

        **NOTE:** If the method :py:meth:``LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string key: key name, which will be tested for presence in properties
        :returns: True, if the specified key is present in properties, False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return key in self._cim_instance

    @lmi_possibly_deleted(None)
    def __getattr__(self, name):
        """
        Returns either a class member, :py:class:`LMIMethod` object, or a
        :py:class:`CIMInstance` object property.

        **NOTE:** If the method :py:meth:``LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string name: class member, the method name, or the property to be returned
        :returns: class member, :py:class:`LMIMethod` object or :py:class:`CIMInstance`
            object property
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        if name in self.__dict__:
            return self.__dict__[name]
        sync_method = False
        if name.startswith("Sync"):
            sync_method = True
            name = name[4:]
        methods = self.methods()
        if methods and name in methods:
            return lmi_wrap_cim_method(self._conn, name, self, sync_method)
        elif name in self._cim_instance:
            member = self._cim_instance[name]
            if isinstance(member, pywbem.CIMInstanceName):
                member = lmi_wrap_cim_instance_name(self._conn, member)
            return member
        raise AttributeError(name)

    @lmi_possibly_deleted(None)
    def __setattr__(self, name, value):
        """
        Modifies a :py:class:`CIMInstance` object property.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string name: name of the :py:class:`CIMInstance`'s member
        :param value: new value
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        if isinstance(value, LMIObjectFactory().LMIInstanceName):
            value = value.wrapped_object
        if name in self._cim_instance:
            t = self._cim_instance.properties[name].type
            self._cim_instance[name] = lmi_cast_to_cim(t, value)
        else:
            self.__dict__[name] = value

    @lmi_possibly_deleted("")
    def __str__(self):
        """
        Returns a string of serialized :py:class:`CIMInstanceName` object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty string. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: string of serialized :py:class:`CIMInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return "%s(classname=\"%s\", %s)" % ( \
            self.__class__.__name__, \
            self.classname, \
            str(self._cim_instance.path) \
        )

    @lmi_possibly_deleted("")
    def __repr__(self):
        """
        Returns a pretty string for the object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty string. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: pretty string for the object
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return "%s(classname=\"%s\", ...)" % (self.__class__.__name__, self.classname)

    def copy(self):
        """
        :returns: copy of itself
        """
        return lmi_wrap_cim_instance(self._conn, self._cim_instance.copy())

    @property
    @lmi_possibly_deleted("")
    def classname(self):
        """
        Property returning a string of a class name.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this method will
        not execute its code and will return an empty string. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: class name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return self._lmi_class.classname

    @property
    @lmi_possibly_deleted("")
    def namespace(self):
        """
        Property retuning a string of a namespace name.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this method will
        not execute its code and will return an empty string. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: namespace name
        :rtype: string
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return self._lmi_class.namespace

    @property
    @lmi_possibly_deleted(None)
    def path(self):
        """
        Property returning a :py:class:`LMIInstanceName` objec5.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`LMIDeletedObjectError` will be raised.

        :returns: :py:class:`LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return lmi_wrap_cim_instance_name(self._conn, self._cim_instance.path)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance)
    def doc(self):
        """
        Prints out pretty verbose message with documentation for the instance. If the LMIShell
        is run in a interactive mode, the output will be redirected to a pager set by
        environment variable :envvar:`PAGER`. If there is not :envvar:`PAGER` set, less or
        more will be used as a fall-back.

        **NOTE:** If the method :py:meth:`LMIInstance.delete` was called, this method will
        not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        LMIXmlFormatter(self._cim_instance.tocimxml()).fancy_format(self._conn._client.interactive)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance)
    def tomof(self):
        """
        Prints out a message with MOF representation of :py:class:`CIMMethod`. If
        the LMIShell is run in a interactive mode, the output will be
        redirected to a pager set by environment variable :envvar:`PAGER`. If
        there is not :envvar:`PAGER` set, less or more will be used as a
        fall-back.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        LMIMofFormatter(self._cim_instance.tomof()).fancy_format(self._conn._client.interactive)

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def associator_names(self, **kwargs):
        """
        Returns a list of associated :py:class:`LMIInstanceName` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return an empty list. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of names by mandating that each returned name identify an
            object that shall be associated to the source object through an instance of
            this class or one of its subclasses. Default value is None.
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of names by mandating that each returned name identify an object
            that shall be either an instance of this class (or one of its subclasses) or
            be this class (or one of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of names by mandating that each returned name identify an object that shall be
            associated to the source object through an association in which the source
            object plays the specified role. That is, the name of the property in the
            association class that refers to the source object shall match the value of
            this parameter. Default value is None.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of names by mandating that each returned name identify an object that
            shall be associated to the source object through an association in which the
            named returned object plays the specified role. That is, the name of the
            property in the association class that refers to the returned object shall
            match the value of this parameter. Default value is None.
        :returns: list of associated :py:class:`LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instance_names`.
        """
        assoc_names_list = self._conn._client._get_associator_names(self._cim_instance, **kwargs)
        return map(lambda assoc_name: lmi_wrap_cim_instance_name(self._conn, assoc_name), assoc_names_list)

    @lmi_possibly_deleted(None)
    def first_associator_name(self, **kwargs):
        """
        Returns the first associated :py:class:`LMIInstanceName` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of names by mandating that each returned name identify an
            object that shall be associated to the source object through an instance of
            this class or one of its subclasses. Default value is None.
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of names by mandating that each returned name identify an object
            that shall be either an instance of this class (or one of its subclasses) or
            be this class (or one of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of names by mandating that each returned name identify an object that shall be
            associated to the source object through an association in which the source
            object plays the specified role. That is, the name of the property in the
            association class that refers to the source object shall match the value of
            this parameter. Default value is None.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of names by mandating that each returned name identify an object that
            shall be associated to the source object through an association in which the
            named returned object plays the specified role. That is, the name of the
            property in the association class that refers to the returned object shall
            match the value of this parameter. Default value is None.
        :returns: first associated :py:class:`LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instance_names`.
        """
        result = self.associator_names(**kwargs)
        if not result:
            return None
        return result[0]

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def associators(self, **kwargs):
        """
        Returns a list of associated :py:class:`LMIInstance` objects with this instance.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return an empty list. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of objects by mandating that each returned object shall be
            associated to the source object through an instance of this class or one of
            its subclasses. Default value is None.
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be either
            an instance of this class (or one of its subclasses) or be this class (or one
            of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall be associated with the
            source object through an association in which the source object plays the
            specified role. That is, the name of the property in the association class
            that refers to the source object shall match the value of this parameter.
            Default value is None.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of objects by mandating that each returned object shall be associated to
            the source object through an association in which the returned object plays
            the specified role. That is, the name of the property in the association class
            that refers to the returned object shall match the value of this parameter.
            Default value is None.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers for each
            object (including qualifiers on the object and on any returned properties)
            shall be included as ``<QUALIFIER>`` elements in the response. Default value
            is False.
        :param bool IncludeClassOrigin: bool flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each returned
            object. Default value is False.
        :param list PropertyList: if not None, the members of the array define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If *PropertyList* is an empty list, no
            properties are included in each returned object. If it is None, no additional
            filtering is defined. Default value is None.
        :returns: list of associated :py:class:`LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instances`.
        """
        associators_list = self._conn._client._get_associators(self._cim_instance, **kwargs)
        return map(lambda assoc: lmi_wrap_cim_instance(self._conn, assoc, assoc.classname, \
            assoc.path.namespace), associators_list)

    @lmi_possibly_deleted(None)
    def first_associator(self, **kwargs):
        """
        Returns the first associated :py:class:`LMIInstance` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of objects by mandating that each returned object shall be
            associated to the source object through an instance of this class or one of
            its subclasses. Default value is None.
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be either
            an instance of this class (or one of its subclasses) or be this class (or one
            of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall be associated with the
            source object through an association in which the source object plays the
            specified role. That is, the name of the property in the association class
            that refers to the source object shall match the value of this parameter.
            Default value is None.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of objects by mandating that each returned object shall be associated to
            the source object through an association in which the returned object plays
            the specified role. That is, the name of the property in the association class
            that refers to the returned object shall match the value of this parameter.
            Default value is None.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers for each
            object (including qualifiers on the object and on any returned properties)
            shall be included as ``<QUALIFIER>`` elements in the response. Default value
            is False.
        :param bool IncludeClassOrigin: bool flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each returned
            object. Default value is False.
        :param list PropertyList: if not None, the members of the array define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If PropertyList is an empty list, no
            properties are included in each returned object. If it is None, no additional
            filtering is defined. Default value is None.
        :returns: first associated :py:class:`LMIInstance`
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`associators_instances`.
        """
        result = self.associators(**kwargs)
        if not result:
            return None
        return result[0]

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def reference_names(self, **kwargs):
        """
        Returns a list of association :py:class:`LMIInstanceName` objects with this
        object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return an empty list. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of object names by mandating that each returned Object Name
            identify an instance of this class (or one of its subclasses) or this class
            (or one of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of object names by mandating that each returned object name shall identify an
            object that refers to the target instance through a property with a name that
            matches the value of this parameter. Default value is None.
        :returns: list of association :py:class:`LMIInstanceName` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instance_names`.
        """
        reference_names_list = self._conn._client._get_reference_names(self._cim_instance, **kwargs)
        return map(lambda ref_name: lmi_wrap_cim_instance_name(self._conn, ref_name), reference_names_list)

    @lmi_possibly_deleted(None)
    def first_reference_name(self, **kwargs):
        """
        Returns the first association :py:class:`LMIInstanceName` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of object names by mandating that each returned Object Name
            identify an instance of this class (or one of its subclasses) or this class
            (or one of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of object names by mandating that each returned object name shall identify an
            object that refers to the target instance through a property with a name that
            matches the value of this parameter. Default value is None.
        :returns: first association :py:class:`LMIInstanceName` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instance_names`.
        """
        result = self.reference_names(**kwargs)
        if not result:
            return None
        return result[0]

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.path, [])
    def references(self, **kwargs):
        """
        Returns a list of association :py:class:`LMIInstance` objects with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty list. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be an
            instance of this class (or one of its subclasses) or this class (or one of its
            subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall refer to the target
            object through a property with a name that matches the value of this
            parameter. Default value is None.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for each object
            (including qualifiers on the object and on any returned properties) shall be
            included as ``<QUALIFIER>`` elements in the response. Default value is False.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN`` attribute
            shall be present on all appropriate elements in each returned object. Default
            value is False.
        :param list PropertyList: if not None, the members of the list define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If PropertyList is an empty list, no
            properties are included in each returned object. If PropertyList is None, no
            additional filtering is defined. Default value is None.
        :returns: list of association :py:class:`LMIInstance` objects
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instances`.
        """
        references_list = self._conn._client._get_references(self._cim_instance, **kwargs)
        return map(lambda ref: lmi_wrap_cim_instance(self._conn, ref, ref.classname, \
            ref.path.namespace), references_list)

    @lmi_possibly_deleted(None)
    def first_reference(self, **kwargs):
        """
        Returns the first association :py:class:`LMIInstance` with this object.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method will
        not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be an
            instance of this class (or one of its subclasses) or this class (or one of its
            subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall refer to the target
            object through a property with a name that matches the value of this
            parameter. Default value is None.
        :param bool IncludeQualifiers: flag indicating, if all qualifiers for each object
            (including qualifiers on the object and on any returned properties) shall be
            included as ``<QUALIFIER>`` elements in the response. Default value is False.
        :param bool IncludeClassOrigin: flag indicating, if the ``CLASSORIGIN`` attribute
            shall be present on all appropriate elements in each returned object. Default
            value is False.
        :param list PropertyList: if not None, the members of the list define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If PropertyList is an empty list, no
            properties are included in each returned object. If PropertyList is None, no
            additional filtering is defined. Default value is None.
        :returns: first association :py:class:`LMIInstance` object
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`references_instances`.
        """
        result = self.references(**kwargs)
        if not result:
            return None
        return result[0]

    @lmi_possibly_deleted([])
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.properties, [])
    def properties(self):
        """
        Returns a list of :py:class:`CIMInstance` properties.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty list. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: list of :py:class:`CIMInstance` properties
        :rtype: list
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        return self._cim_instance.properties.keys()

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance.properties)
    def print_properties(self):
        """
        Prints out the list of :py:class:`CIMInstance` properties.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        for (name, prop) in self._cim_instance.properties.iteritems():
            sys.stdout.write("%s\n" % name)

    @lmi_possibly_deleted({})
    @lmi_return_val_if_fail(lambda obj: obj._cim_instance.properties, {})
    def properties_dict(self):
        """
        Returns dictionary containing property name and value pairs.
        This method may consume significant memory amount when called.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty dictionary. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: dictionary of :py:class:`CIMInstance` properties
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return pywbem.NocaseDict({
            k: x.value for k, x in self._cim_instance.properties.iteritems()})

    @lmi_possibly_deleted(None)
    def property_value(self, prop_name):
        """
        Returns a :py:class:`CIMInstance` property value.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :param string prop_name: :py:class:`CIMInstance` property name
        :raises: :py:exc:`.LMIDeletedObjectError`
        """
        return getattr(self, prop_name)

    @lmi_possibly_deleted([])
    def methods(self):
        """
        Returns a list of :py:class:`CIMInstance` methods' names.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return an empty list. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: list of :py:class:`CIMInstance` methods' names
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_methods`.
        """
        return self._lmi_class.methods()

    @lmi_possibly_deleted(None)
    def print_methods(self):
        """
        Prints out the list of :py:class:`CIMInstance` methods' names.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_methods`.
        """
        self._lmi_class.print_methods()

    @lmi_possibly_deleted( \
        LMIReturnValue(rval=False, errorstr="This instance has been deleted from a CIM broker"), \
    )
    @lmi_return_val_if_fail( \
        lambda obj: obj._cim_instance,
        LMIReturnValue(rval=False, errorstr="Can not refresh the instance"), \
    )
    def refresh(self):
        """
        Retrieves a new :py:class:`CIMInstance` object. Basically refreshes the object
        properties. Returns LMIReturnValue with rval set to True, if the wrapped
        :py:class:`CIMInstance` object was refreshed; otherwise rval is set to False.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return :py:class:`LMIReturnValue` object
        containing False as a return value with proper error string set. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to True, if
            refreshed; False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_refreshing`.
        """
        (new_cim_instance, _, errorstr) = self._conn._client._get_instance(self.path, LocalOnly=False)
        if not new_cim_instance:
            return LMIReturnValue(rval=False, errorstr=errorstr)
        self._cim_instance = new_cim_instance
        return LMIReturnValue(rval=True)

    @lmi_possibly_deleted( \
        LMIReturnValue(rval=False, errorstr="This instance has been deleted from a CIM broker") \
    )
    @lmi_return_val_if_fail( \
        lambda obj: obj._cim_instance.path, \
        LMIReturnValue(rval=False, errorstr="Can't push this instance"), \
    )
    def push(self):
        """
        Pushes the modified object to the CIMOM.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return :py:class:`LMIReturnValue` object
        containing False as a return value with proper error string set. If the shell uses
        exceptions, :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to 0, if modified;
            -1 otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_properties`.
        """
        (rval, rparams, errorstr) = self._conn._client._modify_instance(self._cim_instance)
        return LMIReturnValue(rval=rval, rparams=rparams, errorstr=errorstr)

    @lmi_possibly_deleted(None)
    @lmi_return_if_fail(lambda obj: obj._cim_instance.path)
    def delete(self):
        """
        Deletes this instance from the CIMOM.

        **NOTE:** If the method :py:meth:`.LMIInstance.delete` was called, this method
        will not execute its code and will return None. If the shell uses exceptions,
        :py:exc:`.LMIDeletedObjectError` will be raised.

        :returns: True, if the instance is deleted; False otherwise
        :raises: :py:exc:`.LMIDeletedObjectError`

        **Usage:** :ref:`instances_delete`.
        """
        (rval, rparams, errorstr) = self._conn._client._delete_instance(self.path)
        self._deleted = rval == 0
        return self._deleted

    @property
    def is_deleted(self):
        """
        :returns: True, if the instance was deleted from the CIMOM; False otherwise
        """
        return self._deleted

    @property
    @lmi_possibly_deleted(None)
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`CIMInstance` object
        """
        return self._cim_instance



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from LMIBaseClient import LMIBaseClient
from LMIReturnValue import LMIReturnValue
from LMIShellCache import LMIShellCache

class LMIShellClient(LMIBaseClient):
    """
    :class:`LMIShellClient` overrides few methods due to caching purposes.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be performed
    :param string password: user's password
    :param bool interactive: flag indicating, if the LMIShell client is running in the
        interactive mode; default value is False.
    :param bool use_cache: flag indicating, if the LMIShell client should use cache for
        :class:`CIMClass` objects. This saves a lot's of communication, if there is often
        the :meth:`LMIShellClient._get_class_names()` or
        :meth:`LMIShellClient.attr._get_class()` call issued.
    :param conn_type: type of connection; can be of 2 values:

      * :attr:`LMIBaseClient.CONN_TYPE_WBEM` -- WBEM connection,
      * :attr:`LMIBaseClient.CONN_TYPE_PEGASUS_UDS` -- applicable only for
        Tog-Pegasus CIMOM, it uses Unix socket for the connection; default value is
        :attr:`LMIBaseClient.CONN_TYPE_WBEM`

    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: indicates, whether a server side certificate needs to
        be verified, if SSL used; default value is True

    **NOTE:** If interactive is set to True, LMIShell will:

    * prompt for username and password, if missing and connection via
      Unix socket can not be established.
    * use pager for the output of: :py:meth:`.LMIInstance.doc`,
      :py:meth:`.LMIClass.doc`, :py:meth:`.LMIInstance.tomof` and
      :py:meth:`.LMIMethod.tomof`
    """
    def __init__(self, uri, username="", password="", **kwargs):
        # Set remaining arguments
        interactive = kwargs.pop("interactive", False)
        use_cache = kwargs.pop("use_cache", True)
        conn_type = kwargs.pop("conn_type", LMIBaseClient.CONN_TYPE_WBEM)
        key_file = kwargs.pop("key_file", None)
        cert_file = kwargs.pop("cert_file", None)
        verify_server_cert = kwargs.pop("verify_server_cert", True)
        if kwargs:
            raise TypeError("__init__() got an unexpected keyword arguments: %s" % ", ".join(kwargs.keys()))

        super(LMIShellClient, self).__init__(uri, username, password, conn_type=conn_type,
            key_file=key_file, cert_file=cert_file, verify_server_cert=verify_server_cert)
        self._interactive = interactive
        self._cache = LMIShellCache(use_cache)

    def _get_class_names(self, namespace=None, **kwargs):
        """
        Returns a list of strings containing class names.

        :param string namespace: namespace, from which the class names list
            should be retrieved; if None, default namespace will be used (**NOTE:** see
            :mod:`pywbem`)
        :param string ClassName: defines the class that is the basis for the enumeration.
            If the ClassName input parameter is absent, this implies that the names of all
            classes. Default value is None.
        :param bool DeepInheritance: if not present, of False, only the names of immediate
            child subclasses are returned, otherwise the names of all subclasses of the
            specified class should be returned. Default value is False.
        :returns: :class:`LMIReturnValue` object with ``rval`` set to list of
            string with class names, if no error occurs; otherwise ``rval`` is
            set to None and ``errorstr`` contains an appropriate error string.
        :raises: :exc:`pywbem.CIMError`,
            :exc:`pywbem.AuthError`
        """
        if self._cache.active:
            class_list = self._cache.get_classes()
            if class_list is None:
                (class_list, _, errorstr) = LMIBaseClient._get_class_names(self, namespace, **kwargs)
                if not class_list:
                    return LMIReturnValue(rval=class_list, errorstr=errorstr)
                self._cache.set_classes(class_list)
            return LMIReturnValue(rval=class_list)
        return LMIBaseClient._get_class_names(self, namespace, **kwargs)

    def _get_class(self, class_name, namespace=None, **kwargs):
        """
        Returns a :class:`CIMClass` object.

        :param string class_name: class name
        :param string namespace: namespace name, from which the class should be retrieved;
            if None, default namespace will be used (**NOTE:** see :mod:`pywbem`)
        :param bool LocalOnly:indicates, if only local members should be present in the
            returned :class:`CIMClass`; any CIM elements (properties, methods, and
            qualifiers), except those added or overridden in the class as specified in the
            classname input parameter, shall not be included in the returned class
        :param bool IncludeQualifiers: indicates, if qualifiers for the class (including
            qualifiers on the class and on any returned properties, methods, or method
            parameters) shall be included in the response
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN`` attribute shall
            be present on all appropriate elements in the returned class
        :param list PropertyList: if present and not None, the members of the list define
            one or more property names. The returned class shall not include elements for
            properties missing from this list. Note that if LocalOnly is specified as
            True, it acts as an additional filter on the set of properties returned. For
            example, if property A is included in the PropertyList but LocalOnly is set to
            True and A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no properties
            are included in the response. If the PropertyList input parameter is None, no
            additional filtering is defined.
        :returns: :class:`LMIReturnValue` object with ``rval`` set to :class:`CIMClass`,
            if no error occurs; otherwise ``rval`` is set to None and ``errorstr`` to
            appropriate error string
        :raises: :exc:`pywbem.CIMError`,
            :exc:`pywbem.AuthError`
        """
        if self._cache.active:
            klass = self._cache.get_class(class_name)
            if klass is None:
                (klass, _, errorstr) = LMIBaseClient._get_class(self, class_name, namespace, **kwargs)
                if not klass:
                    return LMIReturnValue(rval=klass, errorstr=errorstr)
                self._cache.add_class(klass)
                self._cache.add_superclass(klass.classname, klass.superclass, namespace)
            return LMIReturnValue(rval=klass)
        return LMIBaseClient._get_class(self, class_name, namespace, **kwargs)

    def _get_superclass(self, classname, namespace=None):
        """
        Returns string of a superclass to given class, if such superclass exists,
        None otherwise.

        :param string classname: class name
        :param string namespace: namespace name
        :returns: superclass name to a given classname or None
        :raises: :exc:`pywbem.CIMError`,
            :exc:`pywbem.AuthError`
        """
        if self._cache.active:
            if self._cache.has_superclass(classname, namespace):
                superclass = self._cache.get_superclass(classname, namespace)
            else:
                superclass = LMIBaseClient._get_superclass(self, classname, namespace)
                self._cache.add_superclass(classname, superclass, namespace)
            return superclass
        return LMIBaseClient._get_superclass(self, classname, namespace)

    @property
    def interactive(self):
        """
        :returns: flag, if the LMIShell is run in the interactive mode
        :rtype: bool
        """
        return self._interactive

    @interactive.setter
    def interactive(self, i):
        """
        Property setter for interactive flag.

        :param bool i: new :attr:`LMIShellClient.interactive` state
        """
        self._interactive = bool(i)

    @property
    def use_cache(self):
        """
        :returns: flag, which tells, if the LMIShell should use a cache
        :rtype: bool
        """
        return self._cache.active

    @use_cache.setter
    def use_cache(self, active):
        """
        Property setter for cache activity.

        :param bool active: new :attr:LMIShellClient.use_cache` state
        """
        self._cache.active = active



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import rlcompleter

from LMIReturnValue import LMIReturnValue
from LMIConstantValues import LMIConstantValues
from LMINamespace import LMINamespace
from LMINamespace import LMINamespaceRoot
from LMIClass import LMIClass
from LMIInstance import LMIInstance
from LMIInstanceName import LMIInstanceName
from LMIMethod import LMIMethod
from LMIConnection import LMIConnection

class LMICompleter(rlcompleter.Completer):
    """
    This LMIShell completer, which is used in the interactive mode, provides
    tab-completion for user friendliness.

    :param dictionary namespace: dictionary, where to perform a completion. If
        unspecified, the default namespace where completions are performed is
        :samp:`__main__` (technically, :samp:`__main__.__dict__`).
    """
    LMI_CLASSES = (
        LMINamespace,
        LMIClass,
        LMIInstance,
        LMIInstanceName,
        LMIMethod,
        LMIConstantValues,
        LMIReturnValue
    )

    def __init__(self, namespace=None):
        rlcompleter.Completer.__init__(self, namespace)
        self._last_complete = []

    def __complete_object_methods(self, obj, text):
        """
        Returns a list of strings of all object methods.

        Arguments:
            obj -- an object, on which the method completion is performed
            text -- string to be completed
        """
        result = filter(lambda m: not m.startswith("_") \
            and callable(getattr(obj, m)) \
            and m.lower().startswith(text), dir(obj))
        return result

    def __complete_object_properties(self, obj, text):
        """
        :param obj: an object, on which the property completion is performed
        :param string text: string to be completed
        :returns: list of all object properties
        """
        result = filter(lambda m: not m.startswith("_") \
            and not callable(getattr(obj, m)) \
            and m.lower().startswith(text), dir(obj))
        return result

    def _callable_postfix(self, val, word):
        """
        :param val: object, which is checked, if it is callable
        :param string word: modified input string, which has opening parentheses appended
        :returns: string with opening parentheses, if the value provided is callable
        """
        if hasattr(val, '__call__') and not isinstance(val, LMICompleter.LMI_CLASSES):
            word = word + "("
        return word

    def complete(self, text, state):
        """
        :param string text: string to be completed.
        :param state: order number of the completion, see rlcompleter
        :returns: completed string
        """
        if not text:
            return ("\t", None)[state]
        if state > 0:
            return self._last_complete[state]
        self._last_complete = []
        members = text.split(".")
        if members[0] in self.namespace:
            cmd = ".".join(members[0:-1])
            to_complete = members[-1].lower()
            expr = eval(cmd, self.namespace)
            methods = self.__complete_object_methods(expr, to_complete)
            properties = self.__complete_object_properties(expr, to_complete)
            if isinstance(expr, (LMIConnection, LMINamespaceRoot)):
                for n in expr.namespaces:
                    if n.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + n)
                methods = [x for x in methods if x not in expr.namespaces]
            elif isinstance(expr, LMINamespace):
                for c in expr.classes():
                    if c.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + c)
            elif isinstance(expr, LMIClass):
                for v in expr.valuemap_properties():
                    if v.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + v + "Values")
            elif isinstance(expr, LMIInstance):
                expr_methods = expr.methods()
                if expr_methods:
                    if to_complete and to_complete in "syn":
                        self._last_complete.append(cmd + ".Sync")
                    else:
                        sync_prefix = ""
                        if to_complete.startswith("sync"):
                            sync_prefix = "Sync"
                            to_complete = to_complete[4:]
                        for m in expr_methods:
                            if m.lower().startswith(to_complete):
                                self._last_complete.append(cmd + "." + sync_prefix + m + "(")
                        to_complete = sync_prefix + to_complete
                for p in expr.properties():
                    if p.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + p)
            elif isinstance(expr, LMIInstanceName):
                for p in expr.key_properties():
                    if p.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + p)
            elif isinstance(expr, LMIMethod):
                for p in expr.valuemap_parameters():
                    if p.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + p + "Values")
            elif isinstance(expr, LMIConstantValues):
                for v in expr.values():
                    if v.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + v)
            elif isinstance(expr, LMIReturnValue):
                for p in expr.properties():
                    if p.lower().startswith(to_complete):
                        self._last_complete.append(cmd + "." + p)
            self._last_complete.extend(cmd + "." + m + "(" for m in methods)
            self._last_complete.extend(cmd + "." + p for p in properties)
            return self._last_complete[state]
        return rlcompleter.Completer.complete(self, text, state)



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import abc
import sys

from LMIBaseObject import LMIBaseObject
from LMIUtil import lmi_cast_to_lmi

class LMIConstantValues(LMIBaseObject):
    """
    Abstract class for constant value objects.

    :param cim_obj: this object is either of type :py:class:`CIMParameter`,
        :py:class:`CIMProperty` or :py:class:`CIMMethod`. Construction of this
        object requires to have a member ``_cast_type`` to properly cast CIM
        object. When constructing derived objects, make sure, that the
        mentioned member is present before calling this constructor.
    :param cast_type: parameter/property cast type
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, cim_obj, cast_type):
        # Keys can contain various undesirable characters, such as python
        # operators, etc. So we drop them.
        keys = map(lambda v: "".join(c for c in v if c.isalnum()),
            cim_obj.qualifiers["Values"].value)
        values = cim_obj.qualifiers["ValueMap"].value
        self._value_map = {}
        self._value_map_inv = {}
        self._cast_type = cast_type
        # Fill two dictionaries for bidirectional access to constant values.
        for i in range(0, len(keys)):
            try:
                key = keys[i]
                val = lmi_cast_to_lmi(self._cast_type, values[i])
                self._value_map[key] = val
                self._value_map_inv[val] = key
            except ValueError, e:
                # Can not cast such value as interval. Can be found in
                # DMTFReserved, VendorReserved values.
                pass

    def __repr__(self):
        """
        Returns a string of all constant names with corresponding value.

        :returns: pretty string
        """
        result = ""
        for (k, v) in self._value_map.iteritems():
            result += "%s = %s\n" % (k, v)
        return result[:-1]

    def __getattr__(self, name):
        """
        Returns either a member of the class, or a constant value.

        Simplifies the code and constant value can be retrieved by
        :samp:`object.constant_value`.

        :param string name: member to retrieve
        :returns: class member
        """
        if name in self.__dict__:
            return self.__dict__[name]
        if name in self._value_map:
            return self._value_map[name]
        raise AttributeError(name)

    def print_values(self):
        """
        Prints all available constant names.

        **Usage:** :ref:`class_get_valuemap_properties`.
        """
        for k in self._value_map.keys():
            sys.stdout.write("%s\n" % k)

    def values_dict(self):
        """
        :returns: dictionary of constants' names and values
        """
        return self._value_map

    def values(self):
        """
        :returns: list of all available constant values
        """
        return self._value_map.keys()

    def value(self, value_name):
        """
        :param string value_name: constant name
        :returns: constant value

        **Usage:** :ref:`class_get_valuemap_property_value`.
        """
        return getattr(self, value_name)

    def value_name(self, value):
        """
        :param int value: numeric constant value
        :returns: constant value
        :rtype: string

        **Usage:** :ref:`class_get_valuemap_property_name`.
        """
        return self._value_map_inv[value]

class LMIConstantValuesParamProp(LMIConstantValues):
    """
    Derived class used for constant values of :py:class:`CIMProperty` and
    :py:class:`CIMParameter`.

    :param cim_property: :py:class:`CIMProperty` or :py:class:`CIMParameter`
        object. Both objects have necessary member ``type`` which is needed for
        proper casting.
    """
    def __init__(self, cim_property):
        super(LMIConstantValuesParamProp, self).__init__(cim_property, cim_property.type)

class LMIConstantValuesMethodReturnType(LMIConstantValues):
    """
    Derived class used for constant values of :py:class:`CIMMethod`.

    :param CIMMethod cim_method: :py:class:`CIMMethod` object
    """
    def __init__(self, cim_method):
        super(LMIConstantValuesMethodReturnType, self).__init__(cim_method, cim_method.return_type)



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys

import LMIExceptions

from LMIBaseObject import LMIWrapperBaseObject
from LMIBaseClient import LMIBaseClient

from LMIUtil import lmi_wrap_cim_namespace
from LMIUtil import lmi_wrap_cim_class
from LMIUtil import lmi_transform_to_lmi

class LMINamespace(LMIWrapperBaseObject):
    """
    LMI class representing CIM namespace.

    :param LMIConnection conn: connection object
    :param string name: namespace name
    """
    def __init__(self, conn, name):
        super(LMINamespace, self).__init__(conn)
        self._name = name

    def __getattr__(self, name):
        """
        Returns a :py:class:`LMIClass` object, but first it fetches the classes list from the
        CIMOM.

        :param string name: class member, class name
        :returns: class member or :py:class:`LMIClass` object
        """
        if name in self.__dict__:
            return self.__dict__[name]
        (class_list, _, _) = self._conn._client._get_class_names(self._name, DeepInheritance=True)
        if name in class_list:
            return lmi_wrap_cim_class(self._conn, name, self.name)
        if '_' in name:
            raise LMIExceptions.LMIClassNotFound(self.name, name)
        raise LMIExceptions.LMINamespaceNotFound(self.name, name)

    def __repr__(self):
        """
        :returns: pretty string for the object
        """
        return "%s(namespace='%s', ...)" % (self.__class__.__name__, self.name)

    def classes(self, filter_key="", exact_match=False):
        """
        Returns a list of class names.

        :param string filter_key: substring of a class name
        :param bool exact_match: tells, if to search for exact match or substring
        :returns: list of class names

        **Usage:** :ref:`namespaces_available_classes`.
        """
        (class_name_list, _, errorstr) = self._conn._client._get_class_names(
            self._name, DeepInheritance=True)
        if not class_name_list:
            return []
        if filter_key:
            if not exact_match:
                filter_lambda = lambda n: n.lower().find(filter_key.lower()) >= 0
            else:
                filter_lambda = lambda n: n.lower() == filter_key.lower()
            class_name_list = filter(filter_lambda, class_name_list)
        return class_name_list

    def print_classes(self, filter_key="", exact_match=False):
        """
        Prints out a list of classes.

        :param string filter_key: substring of a class name
        :param bool exact_match: tells, if to search for exact match, or to search for a
            matching substring

        **Usage:** :ref:`namespaces_available_classes`.
        """
        for c in self.classes(filter_key, exact_match):
            sys.stdout.write("%s\n" % c)

    def cql(self, query):
        """
        Executes a CQL query and returns a list of :py:class:`LMIInstance` objects.

        :param string query: CQL query to execute
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list
            of :py:class:`LMIInstance` objects

        **Usage:** :ref:`namespaces_queries`.
        """
        (inst_list, _, errorstr) = self._conn._client._exec_query(LMIBaseClient.QUERY_LANG_CQL, query, self._name)
        if not inst_list:
            return []
        return lmi_transform_to_lmi(self._conn, inst_list)

    def wql(self, query):
        """
        Executes a WQL query and returns a list of :py:class:`LMIInstance` objects.

        :param string query: WQL query to execute
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list
            of :py:class:`LMIInstance` objects

        **Usage:** :ref:`namespaces_queries`.
        """
        (inst_list, _, errorstr) = self._conn._client._exec_query(LMIBaseClient.QUERY_LANG_WQL, query, self._name)
        if not inst_list:
            return []
        return lmi_transform_to_lmi(self._conn, inst_list)

    @property
    def name(self):
        """
        :returns: namespace name
        :rtype: string
        """
        return self._name

class LMINamespaceRoot(LMINamespace):
    """
    Derived class for *root* namespace. Object of this class is accessible
    from :py:class:`LMIConnection` object as a hierarchy entry.

    :param LMIConnection conn: connection object
    """
    def __init__(self, conn):
        super(LMINamespaceRoot, self).__init__(conn, "root")

    @property
    def cimv2(self):
        """
        :returns: *root/cimv2* namespace
        :rtype: :py:class:`LMINamespace`
        """
        return lmi_wrap_cim_namespace(self._conn, "root/cimv2")

    @property
    def interop(self):
        """
        :returns: *root/interop* namespace
        :rtype: :py:class:`LMINamespace`
        """
        return lmi_wrap_cim_namespace(self._conn, "root/interop")

    @property
    def PG_InterOp(self):
        """
        :returns: *root/PG_InterIp* namespace
        :rtype: :py:class:`LMINamespace`
        """
        return lmi_wrap_cim_namespace(self._conn, "root/PG_InterOp")

    @property
    def PG_Internal(self):
        """
        :returns: *root/PG_Internal* namespace
        :rtype: :py:class:`LMINamespace`
        """
        return lmi_wrap_cim_namespace(self._conn, "root/PG_Internal")

    @property
    def namespaces(self):
        """
        :returns: list of strings with available namespaces

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        return ["cimv2", "interop", "PG_InterOp", "PG_Internal"]

    def print_namespaces(self):
        """
        Prints out all available namespaces accessible via the namespace `root`.

        **Usage:** :ref:`namespaces_available_namespaces`.
        """
        sys.stdout.write("cimv2\n")
        sys.stdout.write("interop\n")
        sys.stdout.write("PG_InterOp\n")
        sys.stdout.write("PG_Internal\n")



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem

from LMIObjectFactory import LMIObjectFactory
from LMIReturnValue import LMIReturnValue
from LMIUtil import lmi_raise_or_dump_exception

from LMIExceptions import LMIFilterError

class LMIBaseClient(object):
    """
    Base client class for CIMOM communication. It abstracts the :py:mod:`pywbem` dependent
    calls to :py:class:`LMIBaseClient` API.

    :param string uri: URI of the CIMOM
    :param string username: account, under which, the CIM calls will be performed
    :param string password: user's password
    :param conn_type: type of connection; can be of 2 values:

      * :py:attr:`LMIBaseClient.CONN_TYPE_WBEM` -- WBEM connection
      * :py:attr:`LMIBaseClient.CONN_TYPE_PEGASUS_UDS` -- applicable only for
        Tog-Pegasus CIMOM, it uses Unix socket for the connection; default value is
        :py:attr:`LMIbaseClient.CONN_TYPE_WBEM`

    :param string key_file: path to x509 key file; default value is None
    :param string cert_file: path to x509 cert file; default value is None
    :param bool verify_server_cert: indicates, whether a server side certificate needs to
        be verified, if SSL used; default value is True
    """
    QUERY_LANG_CQL = "DMTF:CQL"
    QUERY_LANG_WQL = "WQL"
    CONN_TYPE_WBEM, \
    CONN_TYPE_PEGASUS_UDS = range(2)

    def __init__(self, uri, username="", password="", **kwargs):
        def verify_callback(conn, cert, errno, errdepth, rcode):
            """
            Callback function used to verify the server certificate. It is passed to
            OpenSSL.SSL.set_verify, and is called during the SSL handshake. This function
            returns True, if verification passes and False otherwise.

            :param conn: Connection object
            :param cert: x509 object
            :param int errno: potential error number
            :param int errdepth: error depth
            :param int rcode: return code
            :returns: True, if certificate was successfully verified; False otherwise

            **NOTE:** see :py:class:`pywbem.cim_operations.WBEMConnection`
            """
            return bool(rcode)

        # Set remaining arguments
        conn_type = kwargs.pop("conn_type", LMIBaseClient.CONN_TYPE_WBEM)
        verify_server_cert = kwargs.pop("verify_server_cert", True)
        key_file = kwargs.pop("key_file", None)
        cert_file = kwargs.pop("cert_file", None)
        if kwargs:
            raise TypeError("__init__() got an unexpected keyword arguments: %s" % ", ".join(kwargs.keys()))

        self._uri = uri
        self._username = username
        if not self._uri.startswith("http://") and not self._uri.startswith("https://"):
            self._uri= "https://" + self._uri
        if conn_type == LMIBaseClient.CONN_TYPE_PEGASUS_UDS:
            self._cliconn = pywbem.PegasusUDSConnection()
        else:
            self._cliconn = pywbem.WBEMConnection(self._uri,
                (self._username, password),
                x509={"key_file" : key_file, "cert_file" : cert_file},
                verify_callback=verify_callback if verify_server_cert else None
            )

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def _get_instance_names(self, class_name, namespace=None, inst_filter=None, **kwargs):
        """
        Returns a list of :py:class:`CIMInstanceName` objects.

        :param string class_name: class name
        :param string namespace: namespace name, where the instance names live
        :param dictionary inst_filter: dictionary containing filter values. The key
            corresponds to the primary key of the :py:class:`CIMInstanceName`; value contains
            the filtering value.
        :param dictionary kwargs: supported keyword arguments (these are **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`LMIReturnValue` object with ``rval`` contains a list of
            :py:class:`CIMInstanceName` objects, if no error occurs; otherwise ``rval`` is
            set to None and ``errorstr`` contains appropriate error string
        :raises: :py:exc:`.LMIFilterError`,
            :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        filter_value = ""
        filter_key = "Name"
        if inst_filter is None:
            inst_filter = {}
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        if filter_value:
            inst_filter[filter_key] = filter_value
        try:
            inst_name_list = self._cliconn.EnumerateInstanceNames(class_name, namespace, **kwargs)
            if inst_filter:
                inst_name_list_filtered = []
                for inst_name in inst_name_list:
                    append = True
                    for (filter_key, filter_value) in inst_filter.iteritems():
                        if inst_name[filter_key] != filter_value:
                            append = False
                            break
                    if append:
                        inst_name_list_filtered.append(inst_name)
                inst_name_list = inst_name_list_filtered
        except KeyError, e:
            errorstr = "Can not filter by '%s'" % filter_key
            lmi_raise_or_dump_exception(LMIFilterError(errorstr))
            return LMIReturnValue(rval=None, errorstr=errorstr)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return LMIReturnValue(rval=inst_name_list)

    def _get_instance(self, path, **kwargs):
        """
        Returns a :py:class:`CIMInstance` object.

        :param CIMInstanceName path: path of the object, which is about to be retrieved
        :param bool LocalOnly: indicates if to include the only elements (properties,
            methods, references) overridden or defined in the class
        :param bool IncludeQualifiers: indicates, if all Qualifiers for the class and its
            elements shall be included in the response
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN`` attribute shall
            be present on all appropriate elements in the returned class
        :param list PropertyList: if present and not None, the members of the list define
            one or more property names. The returned class shall not include elements for
            properties missing from this list. Note that if LocalOnly is specified as
            True, it acts as an additional filter on the set of properties returned. For
            example, if property A is included in the PropertyList but LocalOnly is set to
            True and A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no properties
            are included in the response. If the PropertyList input parameter is None, no
            additional filtering is defined.

        :returns: :py:class:`LMIReturnValue` object, where ``rval`` is set to
            :pyclass:`CIMInstance` object, if no error occurs; otherwise ``errorstr`` is set
            to corresponding error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(path, LMIObjectFactory().LMIInstanceName):
            path = path.wrapped_object
        try:
            inst = self._cliconn.GetInstance(path, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return LMIReturnValue(rval=inst)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def _get_instances(self, class_name, namespace=None, inst_filter=None, **kwargs):
        """
        Returns a list of :py:class:`CIMInstance` objects.

        :param string class_name: class name
        :param string namespace: namespace, where the instances live
        :param dictionary inst_filter: dictionary containing filter values. The key
            corresponds to the primary key of the :py:class:`CIMInstanceName`;
            value contains the filtering value.
        :param dictionary kwargs: supported keyword arguments (these are **deprecated**)

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** (*string*) -- filtering value, see above

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list of
            :py:class:`CIMIntance` objects, if no error occurs; otherwise
            ``rval`` is set to None and ``errorstr`` is set to corresponding
            error string.
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        filter_value = ""
        filter_key = "Name"
        if inst_filter is None:
            inst_filter = {}
        if "key" in kwargs:
            filter_key = kwargs["key"]
            kwargs.pop("key")
        elif "Key" in kwargs:
            filter_key = kwargs["Key"]
            kwargs.pop("Key")
        if "value" in kwargs:
            filter_value = kwargs["value"]
            kwargs.pop("value")
        if "Value" in kwargs:
            filter_value = kwargs["Value"]
            kwargs.pop("Value")
        if filter_value:
            inst_filter[filter_key] = filter_value
        query = "select * from %s" % class_name
        if inst_filter:
            more = False
            query += " where"
            for (filter_key, filter_value) in inst_filter.iteritems():
                if more:
                    query += " and"
                quotes = isinstance(filter_value, basestring)
                query += " %s =" % filter_key
                query += " \"%s\"" % filter_value if quotes else " %s" % filter_value
                more = True
        (inst_list, _, errorstr) = self._exec_query(LMIBaseClient.QUERY_LANG_WQL, query, namespace)
        if not inst_list:
            return LMIReturnValue(rval=None, errorstr=errorstr)
        return LMIReturnValue(rval=inst_list)

    def _get_class_names(self, namespace=None, **kwargs):
        """
        Returns a list of class names.

        :param string namespace: namespace, from which the class names list
            should be retrieved; if None, default namespace will be used (**NOTE:** see
            :py:mod:`pywbem`)
        :param string ClassName: defines the class that is the basis for the enumeration.
            If the ClassName input parameter is absent, this implies that the names of all
            classes. Default value is None.
        :param bool DeepInheritance: if not present, of False, only the names of immediate
            child subclasses are returned, otherwise the names of all subclasses of the
            specified class should be returned. Default value is False.
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list of strings
            containing class names, if no error occurs; otherwise ``rval`` is set to None
            and ``errorstr`` contains an appropriate error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            class_name_list = self._cliconn.EnumerateClassNames(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return LMIReturnValue(rval=class_name_list)

    def _get_classes(self, namespace=None, **kwargs):
        """
        Returns a :py:class:`CIMClass` object.

        :param string namespace: namespace, from which the class names list
            should be retrieved; if None, default namespace will be used (**NOTE:** see
            :py:mod:`pywbem`)
        :param string ClassName: defines the class that is the basis for the enumeration.
            If the ClassName input parameter is absent, this implies that the names of all
            classes. Default value is None.
        :param bool DeepInheritance: if not present, of False, only the names of immediate
            child subclasses are returned, otherwise the names of all subclasses of the
            specified class should be returned. Default value is False.
        :param bool LocalOnly: indicates, if any CIM elements (properties, methods, and
            qualifiers) except those added or overridden in the class as specified in the
            classname input parameter shall not be included in the returned class.
        :param bool IncludeQualifiers: indicates, if all qualifiers for each class
            (including qualifiers on the class and on any returned properties, methods, or
            method parameters) shall be included as ``<QUALIFIER>`` elements in the
            response.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN`` attribute shall
            be present on all appropriate elements in each returned class.
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list of
            :py:class:`CIMClass` objects, if no error occurs; otherwise ``rval`` is set to
            None and ``errorstr`` to appropriate error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            class_list = self._cliconn.EnumerateClasses(namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return LMIReturnValue(rval=class_list)

    def _get_class(self, class_name, namespace=None, **kwargs):
        """
        Returns a :py:class:`CIMClass` object.

        :param string class_name: class name
        :param string namespace: -- namespace name, from which the :py:class:`CIMClass`
            should be retrieved; if None, default namespace will be used (**NOTE:** see
            :py:mod:`pywbem`)
        :param bool LocalOnly: indicates, if only local members should be present in the
            returned :py:class:`CIMClass`; any CIM elements (properties, methods, and
            qualifiers), except those added or overridden in the class as specified in the
            classname input parameter, shall not be included in the returned class.
            Default value is True.
        :param bool IncludeQualifiers: indicates, if qualifiers for the class (including
            qualifiers on the class and on any returned properties, methods, or method
            parameters) shall be included in the response. Default value is True.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN`` attribute shall
            be present on all appropriate elements in the returned class. Default value is
            False.
        :param list PropertyList: if present and not None, the members of the list define
            one or more property names. The returned class shall not include elements for
            properties missing from this list. Note that if LocalOnly is specified as
            True, it acts as an additional filter on the set of properties returned. For
            example, if property A is included in the PropertyList but LocalOnly is set to
            True and A is not local to the requested class, it is not included in the
            response. If the PropertyList input parameter is an empty list, no properties
            are included in the response. If the PropertyList input parameter is None, no
            additional filtering is defined. Default value is None.
        :returns: :py:class:`LMIReturnValue` object with rval set to
            :py:class:`CIMClass`, if no error occurs; otherwise rval is set to
            none and errorstr to appropriate error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            klass = self._cliconn.GetClass(class_name, namespace, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return LMIReturnValue(rval=klass)

    def _get_superclass(self, class_name, namespace=None):
        """
        Returns a superclass to given class.

        :param string class_name: class name
        :param string namespace: namespace name
        :returns: superclass to given class, if such superclass exists,
            None otherwise
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        (minimal_class, _, _) = LMIBaseClient._get_class(self, class_name, namespace,
            LocalOnly=True, IncludeQualifiers=False, PropertyList=[])
        if not minimal_class:
            return None
        return minimal_class.superclass

    def _call_method_raw(self, instance, method, **params):
        """
        Executes a method within a given :py:class:`CIMInstance`.

        :param CIMInstance instance: object, on which the method will be issued
        :param string method: string containing a method name
        :param dictionary params: parameters passed to the method call
        :returns: :py:class:`LMIReturnValue` object with rval set to return value of the
            method call, rparams set to returned parameters from the method call, if no
            error occurs; otherwise rval is set to -1 and errorstr to appropriate error
            string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(instance, LMIObjectFactory().LMIInstance):
            instance = instance.wrapped_object
        try:
            (rval, rparams) = self._cliconn.InvokeMethod(method, instance.path, **params)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            errorstr = e.args[1] + ": '" + method + "'"
            return LMIReturnValue(rval=-1, errorstr=errorstr)
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            errorstr = e.args[0] + ": '" + method + "'"
            return LMIReturnValue(rval=-1, errorstr=errorstr)
        return LMIReturnValue(rval=rval, rparams=rparams)

    def _call_method(self, instance, method, **params):
        """
        Executes a method within a given :py:class:`CIMInstance`.

        :param CIMInstance instance: object, on which the method will be issued
        :param string method: a method name
        :param dictionary params: parameters passed to the method call
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to True,
            rparams set to returned parameters from the method call, if no
            error occurs; otherwise ``rval`` is set to False and ``errorstr``
            to appropriate error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        (rval, rparams, errorstr) = self._call_method_raw(instance, method, **params)
        return LMIReturnValue(rval=(rval == 0), rparams=rparams, errorstr=errorstr)

    def _get_associator_names(self, instance, **kwargs):
        """
        Returns a list of associated :py:class:`CIMInstanceName` objects with an input
        instance.

        :param CIMInstance instance: for this object the list of associated
            :py:class:`CIMInstanceName` will be returned
        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of names by mandating that each returned name identify an
            object that shall be associated to the source object through an instance of
            this class or one of its subclasses.
        :param string ResultClass: valid CIM class name. It acts as a filter on x the
            returned set of names by mandating that each returned name identify an object
            that shall be either an instance of this class (or one of its subclasses) or
            be this class (or one of its subclasses).
        :param string Role: valid property name. It acts as a filter on the returned set
            of names by mandating that each returned name identify an object that shall be
            associated to the source object through an association in which the source
            object plays the specified role. That is, the name of the property in the
            association class that refers to the source object shall match the value of
            this parameter.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of names by mandating that each returned name identify an object that
            shall be associated to the source object through an association in which the
            named returned object plays the specified role. That is, the name of the
            property in the association class that refers to the returned object shall
            match the value of this parameter.
        :returns: list of associated :py:class:`CIMInstanceName` objects with
            an input instance, if no error occurs; otherwise en empty list is
            returned
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(instance, LMIObjectFactory().LMIInstance):
            instance = instance.wrapped_object
        try:
            return self._cliconn.AssociatorNames(instance.path, **kwargs)
        except (pywbem.cim_operations.CIMError, pywbem.cim_http.AuthError), e:
            lmi_raise_or_dump_exception(e)
            return []

    def _get_associators(self, instance, **kwargs):
        """
        Returns a list of associated :py:class:`CIMInstance` objects with an
        input instance.

        :param CIMInstance instance: for this object the list of associated
            :py:class:`CIMInstance` objects will be returned
        :param string AssocClass: valid CIM association class name. It acts as a filter on
            the returned set of objects by mandating that each returned object shall be
            associated to the source object through an instance of this class or one of
            its subclasses. Default value is None.
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be either
            an instance of this class (or one of its subclasses) or be this class (or one
            of its subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall be associated with the
            source object through an association in which the source object plays the
            specified role. That is, the name of the property in the association class
            that refers to the source object shall match the value of this parameter.
            Default value is None.
        :param string ResultRole: valid property name. It acts as a filter on the returned
            set of objects by mandating that each returned object shall be associated to
            the source object through an association in which the returned object plays
            the specified role. That is, the name of the property in the association class
            that refers to the returned object shall match the value of this parameter.
            Default value is None.
        :param bool IncludeQualifiers: indicates, if all qualifiers for each object
            (including qualifiers on the object and on any returned properties) shall be
            included as ``<QUALIFIER>`` elements in the response. Default value is False.
        :param bool IncludeClassOrigin: indicates, if the ``CLASSORIGIN`` attribute shall
            be present on all appropriate elements in each returned object. Default value
            is False.
        :param list PropertyList: if not None, the members of the array define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If PropertyList is an empty list, no
            properties are included in each returned object. If it is None, no additional
            filtering is defined. Default value is None.
        :returns: list of associated :py:class:`CIMInstance` objects with an input
            instance, if no error occurs; otherwise an empty list is returned
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(instance, LMIObjectFactory().LMIInstance):
            instance = instance.wrapped_object
        try:
            return self._cliconn.Associators(instance.path, **kwargs)
        except (pywbem.cim_operations.CIMError, pywbem.cim_http.AuthError), e:
            lmi_raise_or_dump_exception(e)
            return []

    def _get_reference_names(self, instance, **kwargs):
        """
        Returns a list of association :py:class:`CIMInstanceName` objects.

        :param CIMInstance instance: for this object the association
            :py:class:`CIMInstanceName` objects will be returned
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of object names by mandating that each returned Object Name identify
            an instance of this class (or one of its subclasses) or this class (or one of its
            subclasses).
        :param string Role: valid property name. It acts as a filter on the returned set
            of object names by mandating that each returned object name shall identify an
            object that refers to the target instance through a property with a name that
            matches the value of this parameter.
        :returns: list of association :py:class:`CIMInstanceName` objects with an input
            instance, if no error occurs; otherwise an empty list is returned
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(instance, LMIObjectFactory().LMIInstance):
            instance = instance.wrapped_object
        try:
            return self._cliconn.ReferenceNames(instance.path, **kwargs)
        except (pywbem.cim_operations.CIMError, pywbem.cim_http.AuthError), e:
            lmi_raise_or_dump_exception(e)
            return []

    def _get_references(self, instance, **kwargs):
        """
        Returns a list of association :py:class:`CIMInstance` objects.

        :param CIMInstance instance: for this object the list of association
            :py:class:`CIMInstance` objects will be returned
        :param string ResultClass: valid CIM class name. It acts as a filter on the
            returned set of objects by mandating that each returned object shall be an
            instance of this class (or one of its subclasses) or this class (or one of its
            subclasses). Default value is None.
        :param string Role: valid property name. It acts as a filter on the returned set
            of objects by mandating that each returned object shall refer to the target
            object through a property with a name that matches the value of this
            parameter. Default value is None.
        :param bool IncludeQualifiers: bool flag indicating, if all qualifiers for each
            object (including qualifiers on the object and on any returned properties)
            shall be included as ``<QUALIFIER>`` elements in the response. Default value
            is False.
        :param bool IncludeClassOrigin: bool flag indicating, if the ``CLASSORIGIN``
            attribute shall be present on all appropriate elements in each returned
            object. Default value is False.
        :param list PropertyList: if not None, the members of the list define one or more
            property names. Each returned object shall not include elements for any
            properties missing from this list. If PropertyList is an empty list, no
            properties are included in each returned object. If PropertyList is None, no
            additional filtering is defined. Default value is None.
        :returns: list of association :py:class:`CIMInstance` objects with an input
            instance, if no error occurs; otherwise an empty list is returned
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        if isinstance(instance, LMIObjectFactory().LMIInstance):
            instance = instance.wrapped_object
        try:
            return self._cliconn.References(instance.path, **kwargs)
        except (pywbem.cim_operations.CIMError, pywbem.cim_http.AuthError), e:
            lmi_raise_or_dump_exception(e)
            return []

    def _create_instance(self, classname, namespace, properties=None, qualifiers=None, property_list=None):
        """
        Creates a new :py:class:`CIMInstance` object.

        :param string classname: class name of a new instance
        :param string namespace: namespace, of the new instance
        :param dictionary properties: property names and values
        :param dictionary qualifiers: qualifier names and values
        :param list property_list: list for property filtering; see
            :py:class:`CIMInstance`
        :returns: new :class:`CIMInstance`, if no error occurs; otherwise None is returned
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        # Create a new dictionaries from the input ones, we do not want to modify user's
        # input variables.
        properties = dict(properties) if not properties is None else {}
        qualifiers = dict(qualifiers) if not qualifiers is None else {}
        cim_instance = pywbem.CIMInstance(classname, properties, qualifiers,
            pywbem.CIMInstanceName(classname, namespace=namespace), property_list)
        try:
            cim_path = self._cliconn.CreateInstance(NewInstance=cim_instance)
        except pywbem.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=None, errorstr=e.args[0])
        return self._get_instance(cim_path, LocalOnly=False)

    def _modify_instance(self, instance, **kwargs):
        """
        Modifies a :py:class:`CIMInstance` object at CIMOM side.

        :param CIMInstance instance: object to be modified
        :param bool IncludeQualifiers: indicates, if the qualifiers are modified as
            specified in `ModifiedInstance`. Default value is True.
        :param list PropertyList: if not None, the members of the list define one or more
            property names. Only properties specified in the PropertyList are modified.
            Properties of the *ModifiedInstance* that are missing from the PropertyList
            are ignored. If the PropertyList is an empty list, no properties are modified.
            If the PropertyList is None, the set of properties to be modified consists of
            those of *ModifiedInstance* with values different from the current values in
            the instance to be modified. Default value is None.
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to 0, if no error
            occurs; otherwise ``rval`` is set to -1 and ``errorstr`` is set to corresponding
            error string.
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            self._cliconn.ModifyInstance(instance, **kwargs)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=e.args[0], errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=-1, errorstr=e.args[0])
        return LMIReturnValue(rval=0)

    def _delete_instance(self, instance):
        """
        Deletes a :py:class:`CIMInstance` from the CIMOM side.

        :param CIMInstance instance: object to be deleted
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to 0, if no error
            occurs; otherwise ``rval`` is set to -1 and ``errorstr`` is set to
            corresponding error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            self._cliconn.DeleteInstance(instance)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=e.args[0], errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=-1, errorstr=e.args[0])
        return LMIReturnValue(rval=0)

    def _exec_query(self, query_lang, query, namespace=None):
        """
        Executes a query and returns a list of :py:class:`CIMInstance` objects.

        :param string query_lang: query language
        :param string query: query to execute
        :param string namespace: target namespace for the query
        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to list of
            :py:class:`CIMInstance` objects, if no error occurs; otherwise ``rval`` is set
            to empty list and ``errorstr`` is set to corresponding error string
        :raises: :py:exc:`pywbem.CIMError`, :py:exc:`pywbem.AuthError`
        """
        try:
            inst_list = self._cliconn.ExecQuery(query_lang, query, namespace)
        except pywbem.cim_operations.CIMError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=[], errorstr=e.args[1])
        except pywbem.cim_http.AuthError, e:
            lmi_raise_or_dump_exception(e)
            return LMIReturnValue(rval=[], errorstr=e.args[0])
        return LMIReturnValue(rval=inst_list)

    @property
    def username(self):
        """
        :returns: user name as a part of provided credentials
        :rtype: string
        """
        return self._username

    @property
    def uri(self):
        """
        :returns: URI of the CIMOM
        :rtype: string
        """
        return self._uri



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import pywbem

from LMIBaseObject import LMIWrapperBaseObject
from LMIBaseClient import LMIBaseClient
from LMIFormatter import LMIXmlFormatter
from LMIConstantValues import LMIConstantValuesParamProp
from LMIObjectFactory import LMIObjectFactory

from LMIExceptions import LMIUnknownPropertyError

from LMIDecorators import lmi_class_fetch_lazy
from LMIDecorators import lmi_return_val_if_fail
from LMIDecorators import lmi_return_if_fail

from LMIUtil import lmi_cast_to_cim
from LMIUtil import lmi_raise_or_dump_exception
from LMIUtil import lmi_wrap_cim_instance
from LMIUtil import lmi_wrap_cim_instance_name

class LMIClass(LMIWrapperBaseObject):
    """
    LMI wrapper class representing :py:class:`CIMClass`.

    :param LMIConnection conn: connection object
    :param LMINamespace namespace: namespace object
    :param string classname: CIM class name
    """
    def __init__(self, conn, namespace, classname):
        # We use __dict__ to avoid recursion potentially caused by
        # combo __setattr__ and __getattr__
        self.__dict__["_namespace"] = namespace
        self.__dict__["_cim_class"] = None
        self.__dict__["_cim_classname"] = classname
        super(LMIClass, self).__init__(conn)

    def __repr__(self):
        """
        :returns: a pretty string for the object
        """
        return "%s(classname='%s', ...)" % (self.__class__.__name__, self.classname)

    @lmi_class_fetch_lazy
    def __getattr__(self, name):
        """
        Returns either a class member, or a constant value.

        Simplifies the code and constant value can be retrieved by
        :samp:`object.constant_value`.

        :param string name: object member or constant values member
        :returns: class member of :py:class:`LMIConstantValuesParamProp` object
        """
        if name in self.__dict__:
            return self.__dict__[name]
        if name.endswith("Values"):
            property_name = name[:-6]
            return LMIConstantValuesParamProp(self._cim_class.properties[property_name])
        raise AttributeError(name)

    @lmi_class_fetch_lazy
    @lmi_return_val_if_fail(lambda obj: obj._namespace, None)
    def create_instance(self, properties=None, qualifiers=None, property_list=None):
        """
        Creates a new :py:class:`CIMInstance` at the server side and returns
        :py:class:`LMIReturnValue` object containing :py:class:`LMIInstance` as a result.

        :param dictionary properties: initial properties with corresponding values
        :param dictionary qualifiers: initial qualifiers
        :param list property_list: list of properties, which should be present in
            :py:class:`LMIInstance` object

        **Usage:** See :ref:`class_create_instance`.
        """
        # No need to copy dictionaries to avoid the variable mix-up, the copying is done in
        # LMIBaseClient._create_instance(), we just pass what we get.
        properties = properties if not properties is None else {}
        qualifiers = qualifiers if not qualifiers is None else {}
        self_properties = self._cim_class.properties
        for (key, value) in properties.iteritems():
            if not key in self._cim_class.properties:
                errorstr = "No such instance property '%s'" % key
                lmi_raise_or_dump_exception(LMIUnknownPropertyError(errorstr))
                return None
            if isinstance(value, LMIObjectFactory().LMIInstanceName):
                value = value.wrapped_object
            t = self._cim_class.properties[key].type
            properties[key] = lmi_cast_to_cim(t, value)
        (cim_instance, rparams, errorstr) = self._conn._client._create_instance(self.classname,
            self.namespace, properties, qualifiers, property_list)
        if not cim_instance:
            return None
        return lmi_wrap_cim_instance(self._conn, cim_instance, cim_instance.classname,
            cim_instance.path.namespace)

    @lmi_class_fetch_lazy
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def doc(self):
        """
        Prints out pretty verbose message with documentation for the class. If the
        LMIShell is run in a interactive mode, the output will be redirected to a pager
        set by environment variable :envvar:`PAGER`. If there is not :envvar:`PAGER` set,
        less or more will be used as a fall-back.
        """
        LMIXmlFormatter(self._cim_class.tocimxml()).fancy_format(self._conn._client.interactive)

    def fetch(self):
        """
        Manually fetches a wrapped :py:class:`CIMClass` object.

        **Usage:** See :ref:`class_fetching_a_class`.
        """
        if self._cim_class or not self._namespace:
            # We already have CIMClass, or we do not know the namespace, from which
            # the class should be fetched.
            return
        (self._cim_class, _, _) = self._conn._client._get_class(self._cim_classname,
            self._namespace.name, LocalOnly=False)
        # Store the constant values as a list. This can consume some time, if computed on demand.
        self._valuemap_properties_list = [k for (k, v) in self._cim_class.properties.iteritems() \
            if "ValueMap" in v.qualifiers]

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def instance_names(self, inst_filter=None, **kwargs):
        """
        Returns a LMIReturnValue containing a list of LMIInstanceNames.

        :param dictionary inst_filter: filter values. The key corresponds to the primary
            key of the :py:class:`CIMInstanceName`; value contains the filtering value
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: :py:class:`LMIReturnValue` object with ``rval`` set to a list of
            :py:class:`LMIInstanceName` objects

        **Usage:** See :ref:`class_get_instance_names` and :ref:`class_instance_filtering`.
        """
        (inst_name_list, _, errorstr) = self._conn._client._get_instance_names(
            self._cim_classname, self._namespace.name, inst_filter, **kwargs)
        if not inst_name_list:
            return []
        return map(lambda inst_name: lmi_wrap_cim_instance_name(self._conn, inst_name), inst_name_list)

    def new_instance_name(self, keybindings):
        """
        Create new :py:class:`LMIInstanceName` object by passing all the
            keys/values of the object.

        :param dictionary keybindings: primary keys of instance name with corresponding
            values
        :returns: new :py:class:`LMIInstanceName` object

        **Usage:** See :ref:`class_new_instance_name`.
        """
        kbs = pywbem.NocaseDict()
        for key, value in keybindings.iteritems():
            if isinstance(value, LMIObjectFactory().LMIInstanceName):
                value = value.wrapped_object
            kbs[key] = value
        cim_inst_name = pywbem.CIMInstanceName(self.classname, kbs,
            namespace=self.namespace)
        return lmi_wrap_cim_instance_name(self._conn, cim_inst_name)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def first_instance_name(self, inst_filter=None, **kwargs):
        """
        Returns the first :py:class:`LMIInstanceName` of the corresponding
        class.

        :param dictionary inst_filter: filter values, where the key corresponds to the
            primary key of :py:class:`CIMInstanceName`; value contains the filtering
            value
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: first :py:class:`LMIInstanceName` object

        **Usage:** See :ref:`class_get_instance_names` and :ref:`class_instance_filtering`.
        """
        inst_name_list = self.instance_names(inst_filter, **kwargs)
        if not inst_name_list:
            return None
        return inst_name_list[0]

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def instances(self, inst_filter=None, **kwargs):
        """
        Returns a list of objects of :py:class:`LMIInstance`.

        :param dictionary inst_filter: filter values, where the key corresponds to the key
            of :py:class:`CIMInstance`; value contains the filtering value
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** (*string*) -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: list of :py:class:`LMIInstance` objects

        **Usage:** See :ref:`class_get_instances` and :ref:`class_instance_filtering`.
        """
        (instance_list, _, errorstr) = self._conn._client._get_instances(
            self._cim_classname, self._namespace.name, inst_filter, **kwargs)
        if not instance_list:
            return []
        return map(lambda instance: lmi_wrap_cim_instance(self._conn,
            instance, instance.classname, instance.path.namespace),
            instance_list)

    # NOTE: usage with Key=something, Value=something is deprecated
    # NOTE: inst_filter is either None or dict
    def first_instance(self, inst_filter=None, **kwargs):
        """
        Returns the first :py:class:`LMIInstance` of the corresponding class.

        :param dictionary inst_filter: filter values, where the key corresponds to the key
            of :py:class:`CIMInstance`; value contains the filtering value.
        :param dictionary kwargs: deprecated keyword arguments

            * **Key** or **key** -- filtering key, see above
            * **Value** or **value** -- filtering value, see above

        :returns: first :py:class:`LMIInstance` object

        **Usage:** See :ref:`class_get_instances` and :ref:`class_instance_filtering`.
        """
        filter_value = "value" in [k.lower() for k in kwargs.keys()]
        if inst_filter or filter_value:
            (instance_list, _, errorstr) = self._conn._client._get_instances(
                self._cim_classname, self._namespace.name, inst_filter, **kwargs)
            if not instance_list:
                return None
            instance = instance_list[0]
            return lmi_wrap_cim_instance(self._conn, instance, instance.classname,
                instance.path.namespace)
        inst_name = self.first_instance_name(inst_filter, **kwargs)
        return inst_name.to_instance() if inst_name else None

    @lmi_class_fetch_lazy
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def valuemap_properties(self):
        """
        :returns: list of strings of the constant names

        **Usage:** :ref:`class_valuemap_properties`.
        """
        return self._valuemap_properties_list

    @lmi_class_fetch_lazy
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_valuemap_properties(self):
        """
        Prints out the list of string of constant names.

        **Usage:** :ref:`class_valuemap_properties`.
        """
        for i in self._valuemap_properties_list:
            sys.stdout.write("%s\n" % i)

    @lmi_class_fetch_lazy
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def properties(self):
        """
        :returns: list of strings of the :py:class:`CIMClass` properties

        **Usage:** See :ref:`class_properties`.
        """
        return self._cim_class.properties.keys()

    @lmi_class_fetch_lazy
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_properties(self):
        """
        Prints out the list of :py:class:`CIMClass` properties.

        **Usage:** See :ref:`class_properties`.
        """
        for prop in self._cim_class.properties.keys():
            sys.stdout.write("%s\n" % prop)

    @lmi_class_fetch_lazy
    @lmi_return_val_if_fail(lambda obj: obj._namespace, [])
    def methods(self):
        """
        :returns: list of strings of :py:class:`CIMClass` methods.

        **Usage:** See :ref:`class_methods`.
        """
        return self._cim_class.methods.keys()

    @lmi_class_fetch_lazy
    @lmi_return_if_fail(lambda obj: obj._namespace)
    def print_methods(self):
        """
        Prints out the list of :py:class:`CIMClass` methods.

        **Usage:** See :ref:`class_methods`.
        """
        for method in self._cim_class.methods.keys():
            sys.stdout.write("%s\n" % method)

    @property
    def classname(self):
        """
        :returns: class name
        :rtype: string
        """
        return self._cim_classname

    @property
    @lmi_return_val_if_fail(lambda obj: obj._namespace, "Unknown")
    def namespace(self):
        """
        :returns: namespace name
        :rtype: string
        """
        return self._namespace.name

    @property
    @lmi_class_fetch_lazy
    def wrapped_object(self):
        """
        :returns: wrapped :py:class:`CIMClass` object
        """
        return self._cim_class



# Copyright (C) 2012-2013 Peter Hatina <phatina@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

class LMIIndication(object):
    """
    Class representing a received indication.

    :param tuple export_methods: tuple tree for export methods

    **NOTE:** see :py:mod:`pywbem`
    """
    def __init__(self, export_methods):
        self._export_methods = export_methods

    def export_method_names(self):
        """
        :returns: tuple tree for export methods

        **NOTE:** see :py:mod:`pywbem`
        """
        return self._export_methods.keys()

    def exported_objects(self):
        """
        :returns: list of exported :py:class:`CIMInstance` objects
        """
        result = []
        for m in self._export_methods.values():
            for obj in m.values():
                result.append(obj)
        return result
